# flake8: noqa

import numpy
import numpy as np
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim
from skimage.util import random_noise

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import newyork, pollen, normalise, lizard, dots
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log


def demo(image, name):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """

    # Log.set_log_max_depth(5)

    image = normalise(image.astype(np.float32))
    # noisy = add_noise(image, intensity=None, variance=0.1, sap=0, clip=False)
    noisy = random_noise(image, mode="gaussian", var=0.1, seed=0, clip=False)
    noisier = random_noise(noisy, mode="gaussian", var=0.1, seed=100, clip=False)

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
    )
    regressor = CBRegressor(
        patience=16, loss='l1', learning_rate=0.002, max_num_estimators=4096
    )

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    it.train(noisy, noisy, jinv=True)
    n2s_denoised = it.translate(noisy)

    it.exclude_center_feature = False
    it.train(noisier, noisy, jinv=False)
    denoised = it.translate(noisy)
    denoised_corrected = 2 * denoised - noisy

    # denoised2 = it.translate(it.translate(it.translate(denoised)))
    denoised2 = it.translate(denoised)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    n2s_denoised = numpy.clip(n2s_denoised, 0, 1)
    denoised_corrected = numpy.clip(denoised_corrected, 0, 1)
    denoised2 = numpy.clip(denoised2, 0, 1)

    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)

    psnr_n2s_denoised = psnr(image, n2s_denoised)
    ssim_n2s_denoised = ssim(image, n2s_denoised)

    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)

    psnr_denoised_corrected = psnr(image, denoised_corrected)
    ssim_denoised_corrected = ssim(image, denoised_corrected)

    psnr_denoised2 = psnr(image, denoised2)
    ssim_denoised2 = ssim(image, denoised2)

    print("noisy                             :", psnr_noisy, ssim_noisy)
    print(
        "denoised (classic_denoisers)                    :",
        psnr_n2s_denoised,
        ssim_n2s_denoised,
    )
    print("denoised (noiser2noise)           :", psnr_denoised, ssim_denoised)
    print(
        "denoised (noiser2noise corrected) :",
        psnr_denoised_corrected,
        ssim_denoised_corrected,
    )
    print("denoised (x2)                     :", psnr_denoised2, ssim_denoised2)

    Log.enable_output = False
    denoised_images = []
    for i in range(1, 32):
        psnr_denoised = psnr(image, numpy.clip(denoised, 0, 1))
        ssim_denoised = ssim(image, numpy.clip(denoised, 0, 1))
        psnr_sslos = psnr(numpy.clip(n2s_denoised, 0, 1), numpy.clip(denoised, 0, 1))
        ssim_sslos = ssim(numpy.clip(n2s_denoised, 0, 1), numpy.clip(denoised, 0, 1))
        print(f"x{i} :", psnr_sslos, ssim_sslos, psnr_denoised, ssim_denoised)

        denoised_images.append(numpy.clip(denoised, 0, 1))
        denoised = it.translate(denoised)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(noisier, name='noisier')
        viewer.add_image(n2s_denoised, name='denoised (classic_denoisers)')
        viewer.add_image(denoised, name='denoised (noiser3noise)')
        viewer.add_image(denoised_corrected, name='denoised (noiser3noise corrected)')
        viewer.add_image(numpy.stack(denoised_images), name=f'denoised images')


camera_image = camera()
demo(camera_image, "camera")
lizard_image = lizard()
demo(lizard_image, "lizard")
pollen_image = pollen()
demo(pollen_image, "pollen")
dots_image = dots()
demo(dots_image, "dots")
newyork_image = newyork()
demo(newyork_image, "newyork")



import numpy
from scipy.ndimage import gaussian_filter

from aydin.util.denoise_nd.denoise_nd import extend_nd


def demo_denoise_nd():
    # raw function that only supports 2D images:
    def function(image, sigma):
        if image.ndim != 2:
            raise RuntimeError("Function only supports arrays of dimensions 2")
        return gaussian_filter(image, sigma)

    # extended function that supports all dimension (with all caveats associated to how we actually do this extension...)
    @extend_nd(available_dims=[2])
    def extended_function(image, sigma):
        return function(image, sigma)

    image = numpy.zeros((32, 5, 64))
    image[16, 2, 32] = 1

    denoised = extended_function(image, sigma=1)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(denoised, name='denoised')


demo_denoise_nd()



# flake8: noqa
import numpy
import numpy as np
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import (
    normalise,
    add_noise,
    dots,
    lizard,
    pollen,
    newyork,
    characters,
)
from aydin.it.classic_denoisers.harmonic import calibrate_denoise_harmonic
from aydin.util.log.log import Log


def demo_harmonic(image, display=True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True
    Log.set_log_max_depth(5)

    image = normalise(image.astype(np.float32))
    noisy = add_noise(image)

    function, parameters, memreq = calibrate_denoise_harmonic(noisy)
    denoised = function(noisy, **parameters)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)
    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)
    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)
    print("         noisy   :", psnr_noisy, ssim_noisy)
    print("harmonic denoised:", psnr_denoised, ssim_denoised)

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='image')
            viewer.add_image(noisy, name='noisy')
            viewer.add_image(denoised, name='denoised')

    return ssim_denoised


if __name__ == "__main__":

    demo_harmonic(characters())
    demo_harmonic(pollen())
    demo_harmonic(lizard())
    demo_harmonic(dots())
    demo_harmonic(camera())
    demo_harmonic(newyork())



from typing import Optional
import numpy
from scipy.ndimage import median_filter, gaussian_filter
from scipy.signal import medfilt2d

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart


def calibrate_denoise_gm(
    image,
    max_filter_size: int = 3,
    crop_size_in_voxels: Optional[int] = 128000,
    max_num_evaluations: int = 256,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the Gaussian-Median mix denoiser for the given image and returns
    the optimal parameters obtained using the N2S loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate denoiser for.

    max_filter_size : int
        Max filter size to use during calibration.
        Should be a positive odd number such as 3, 5, 7, ...

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate denoiser.
        (advanced)

    max_num_evaluations: int
        Maximum number of evaluations for finding the optimal parameters.
        (advanced)

    display_images: bool
        When True the denoised images encountered during optimisation are shown

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # Size range:
    filter_size_range = [3, 5, 7]

    # filter sizes:
    filter_size_range = list((s for s in filter_size_range if s <= max_filter_size))

    # Sigma range:
    sigma_range = (0.01, 3.0)

    # factor range:
    factor_range = (1.0, 3.0)

    # Alpha range:
    mixing_range = (0.0, 1.0)

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {
        'size': filter_size_range,
        'sigma': sigma_range,
        'factor': factor_range,
        'alpha': mixing_range,
        'beta': mixing_range,
        'gamma': mixing_range,
    }

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_smart(
            crop,
            denoise_gm,
            denoise_parameters=parameter_ranges,
            display_images=display_images,
            max_num_evaluations=max_num_evaluations,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = 3 * image.nbytes

    return denoise_gm, best_parameters, memory_needed


def denoise_gm(
    image,
    sigma: float = 0.5,
    size: int = 3,
    factor: float = 2,
    alpha: float = 0.25,
    beta: float = 0.3,
    gamma: float = 0.5,
    **kwargs,
):
    """
    Denoises the given image with a linear mixing of median
    filtering and Gaussian filtering. Simple and fast but quite
    effective for low to moderate noise levels and images with
    band-limited signal (~ there is a 'PSF').
    \n\n
    Note: We recommend applying a variance stabilisation transform
    to improve results for images with non-Gaussian noise.

    Parameters
    ----------
    image: ArrayLike
        Image to be denoised.

    sigma: float
        Gaussian blur sigma.

    size: int
        Size of the median filter

    factor: float
        Ratio between the scales of the two scales

    alpha: float
        First mixing coefficient.

    beta: float
        First mixing coefficient.

    gamma: float
        First mixing coefficient.


    Returns
    -------
    Denoised image
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    size_1 = size
    size_2 = int(factor * size / 2) * 2 + 1
    sigma_1 = sigma
    sigma_2 = factor * sigma

    wm1 = alpha * beta
    wm2 = alpha * (1 - beta)
    wg1 = (1 - alpha) * gamma
    wg2 = (1 - alpha) * (1 - gamma)

    if image.ndim == 2:
        denoised = wm1 * medfilt2d(image, kernel_size=size_1)
    else:
        denoised = wm1 * median_filter(image, size=size_1)

    if image.ndim == 2:
        denoised += wm2 * medfilt2d(image, kernel_size=size_2)
    else:
        denoised += wm2 * median_filter(image, size=size_2)

    denoised += wg1 * gaussian_filter(image, sigma=sigma_1)
    denoised += wg2 * gaussian_filter(image, sigma=sigma_2)

    return denoised



from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.keras.layers import multiply


class Maskout(Layer):
    """
    A layer that mutiply mask with image. This is for masking architecture.

    Parameters
    ----------
    kwargs
    """

    def __init__(self, **kwargs):
        super(Maskout, self).__init__(**kwargs)

    def build(self, input_shape):
        """
        Parameters
        ----------
        input_shape : tuple

        """
        assert isinstance(input_shape, list)
        super(Maskout, self).build(input_shape)

    def call(self, x, training=None):
        """

        Parameters
        ----------
        x
        training

        Returns
        -------
        multiply(x)

        """
        assert isinstance(x, list)
        return multiply(x)
        # return K.in_train_phase(multiply(x), x[0], training=training)

    def compute_output_shape(self, input_shape):
        """
        Parameters
        ----------
        input_shape : tuple

        Returns
        -------
        output_shape : tuple

        """
        assert isinstance(input_shape, list)
        shape_a, shape_b = input_shape
        return shape_a



# flake8: noqa

import numpy
import pytest
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import normalise, add_noise, add_blur_2d, cropped_newyork
from aydin.it.deconvolution.llr_deconv import ImageTranslatorLearnedLRDeconv
from aydin.it.deconvolution.lr_deconv_scipy import ImageTranslatorLRDeconvScipy
from aydin.util.log.log import Log


@pytest.mark.heavy
def test():
    image = cropped_newyork()

    Log.enable_output = True

    image = normalise(image.astype(numpy.float32))
    blurred_image, psf_kernel = add_blur_2d(image)
    noisy_and_blurred_image = add_noise(
        blurred_image, intensity=20, variance=0.005, sap=0.001
    )

    lr = ImageTranslatorLearnedLRDeconv(
        psf_kernel=psf_kernel,
        max_num_iterations=30,
        deconvolution_class=ImageTranslatorLRDeconvScipy,
    )

    lr.train(noisy_and_blurred_image, noisy_and_blurred_image)
    lr_deconvolved_image = lr.deconvolve(noisy_and_blurred_image)
    llr_deconvolved_image = lr.translate(noisy_and_blurred_image)

    assert ssim(image, noisy_and_blurred_image) < 0.25
    assert ssim(image, lr_deconvolved_image) > 0.20
    assert ssim(image, llr_deconvolved_image) > 0.44


# image = newyork()
#
# demo(image)



import numpy
from skimage.exposure import rescale_intensity

from aydin.features.groups.uniform import UniformFeatures
from aydin.io.datasets import camera


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def test_uniform_feature_group():
    # get image:
    image = n(camera().astype(numpy.float32))

    # Instantiates median features:
    uniform = UniformFeatures(
        include_corner_features=True, include_scale_one=True, include_fine_features=True
    )

    for description in uniform._get_feature_descriptions_list(image.ndim):
        print(description)
    assert uniform.num_features(image.ndim) == 65

    # Check receptive field radius:
    assert uniform.receptive_field_radius == 3070

    # Set image:
    uniform.prepare(image)

    # compute features and check their valididty:
    feature = numpy.empty_like(image)
    for index in range(uniform.num_features(image.ndim)):
        uniform.compute_feature(index=index, feature=feature)
        assert (feature != image).any()



import numpy
from aydin.nn.util.mask_generator import maskedgen, randmaskgen


def test_checkerbox_maskgen():
    image = numpy.ones((20, 64, 64, 1))
    batch_size = 10
    mask_size = (3, 3)

    maskgen = maskedgen(image, batch_size, mask_size, replace_by='zero')

    minibatch = next(maskgen)
    actual_mask_ratio = minibatch[0]['input_msk'].sum() / minibatch[0]['input_msk'].size

    assert len(minibatch) == 2
    assert isinstance(minibatch[0], dict)
    assert (
        minibatch[0]['input'].sum()
        == minibatch[0]['input_msk'].size - minibatch[0]['input_msk'].sum()
    )
    assert minibatch[0]['input_msk'].sum() == minibatch[1].sum()
    assert minibatch[1].shape == (batch_size,) + image.shape[1:]
    assert (
        (1 / numpy.prod(mask_size)) * 0.9
        <= actual_mask_ratio
        <= (1 / numpy.prod(mask_size)) * 1.1
    )


def test_random_maskgen():
    image = numpy.ones((20, 64, 64, 1))
    batch_size = 10
    p_maskedpixels = 0.1

    maskgen = randmaskgen(
        image, batch_size, p_maskedpixels=p_maskedpixels, replace_by='zero'
    )

    minibatch = next(maskgen)
    actual_mask_ratio = minibatch[0]['input_msk'].sum() / minibatch[0]['input_msk'].size

    assert len(minibatch) == 2
    assert isinstance(minibatch[0], dict)
    assert (
        minibatch[0]['input'].sum()
        == minibatch[0]['input_msk'].size - minibatch[0]['input_msk'].sum()
    )
    assert minibatch[0]['input_msk'].sum() == minibatch[1].sum()
    assert minibatch[1].shape == (batch_size,) + image.shape[1:]
    assert p_maskedpixels * 0.9 <= actual_mask_ratio <= p_maskedpixels * 1.1



from pprint import pprint

import numpy
from skimage.util import random_noise

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, pollen
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.motion import MotionStabilisationTransform
from aydin.it.transforms.range import RangeTransform
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import Log


def demo_motion():

    Log.enable_output = True

    shifts = tuple((5 * i, int(0.5 * i * i)) for i in range(10))

    print('')
    pprint(shifts)

    image = normalise(pollen())[0:256, 0:256]
    array = numpy.stack(
        [add_noise(numpy.roll(image, shift=shift, axis=(0, 1))) for shift in shifts]
    )

    mc = MotionStabilisationTransform(axes=0)

    preprocessed_array = mc.preprocess(array.copy())
    postprocessed_array = mc.postprocess(preprocessed_array.copy())

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(array, name='array')
        viewer.add_image(preprocessed_array, name='corrected_array')
        viewer.add_image(postprocessed_array, name='uncorrected_array')

    # assert not (array == corrected_array).all()
    assert (array == postprocessed_array).all()

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
    )
    regressor = LGBMRegressor(patience=128, compute_training_loss=True)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)
    it.add_transform(RangeTransform())
    it.train(array)
    denoised_without_preprocessing = it.translate(array)

    it.add_transform.append(mc)
    it.train(array)
    denoised_with_preprocessing = it.translate(array)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(array, name='array')
        viewer.add_image(preprocessed_array, name='preprocessed_array')
        viewer.add_image(postprocessed_array, name='postprocessed_array')
        viewer.add_image(
            denoised_without_preprocessing, name='denoised_without_preprocessing'
        )
        viewer.add_image(
            denoised_with_preprocessing, name='denoised_with_preprocessing'
        )


def add_noise(image, intensity=4, variance=0.4):
    noisy = image
    if intensity is not None:
        noisy = numpy.random.poisson(image * intensity) / intensity
    noisy = random_noise(noisy, mode="gaussian", var=variance)
    noisy = noisy.astype(numpy.float32, copy=False)
    return noisy


demo_motion()



import numpy
from skimage.data import binary_blobs

from aydin.io.datasets import normalise
from aydin.it.transforms.padding import PaddingTransform


def test_padding():
    image = binary_blobs(length=128, seed=1, n_dim=3).astype(numpy.float32)
    image = normalise(image)

    pt = PaddingTransform(pad_width=17)

    preprocessed = pt.preprocess(image)
    postprocessed = pt.postprocess(preprocessed)

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(attenuated, name='attenuated')
    #     viewer.add_image(preprocessed, name='preprocessed')
    #     viewer.add_image(postprocessed, name='postprocessed')

    assert postprocessed.dtype == image.dtype
    assert postprocessed.shape == image.shape
    assert numpy.abs(postprocessed - image).mean() < 1e-8



import numpy

from aydin.io.datasets import newyork
from aydin.it.transforms.range import RangeTransform


def test_range_minmax():
    do_test_range("minmax")


def test_range_percentile():
    do_test_range("percentile")


def do_test_range(mode):
    image = newyork()

    rt = RangeTransform(mode=mode)

    preprocessed = rt.preprocess(image)
    postprocessed = rt.postprocess(preprocessed)

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(preprocessed, name='preprocessed')
    #     viewer.add_image(postprocessed, name='postprocessed')

    assert postprocessed.dtype == image.dtype
    assert postprocessed.shape == image.shape
    assert numpy.abs(postprocessed - image).mean() < 1e-8



import warnings
import numpy as np
from tensorflow.python.keras import backend
from tensorflow.python.keras.callbacks import Callback

from aydin.util.log.log import lprint


class CNNCallback(Callback):
    """
    tensorflow.keras Callback to linkup to the it callback machinery

    Parameters
    ----------
    monitor_callback
    """

    def __init__(self, monitor_callback):
        super().__init__()

        self.monitor_callback = monitor_callback
        self.iteration = 0

    def on_train_begin(self, logs=None):
        pass

    def on_batch_ends(self, batch, logs=None):
        self.notify(logs)

    def on_epoch_end(self, epoch, logs=None):
        self.notify(logs)

    def on_train_end(self, logs=None):
        self.notify(logs)

    def notify(self, logs):
        if self.monitor_callback:
            iteration = self.iteration
            val_loss = self.get_monitor_value(logs)
            self.monitor_callback(iteration, val_loss)
            self.iteration += 1
        pass

    def get_monitor_value(self, logs):
        if logs:
            if logs.get('val_loss'):
                monitor_value = logs.get('val_loss')
            else:
                monitor_value = logs.get('loss')
        else:
            monitor_value = -1
            # monitor_value = self.params['metrics']
        return monitor_value


class EarlyStopping(Callback):
    """
    Stop training when a monitored quantity has stopped improving.

    Parameters
    ----------
    parent
    monitor
        quantity to be monitored
    min_delta
        minimum change in the monitored quantity
        to qualify as an improvement, i.e. an absolute
        change of less than min_delta, will count as no
        improvement
    patience
        number of epochs with no improvement
        after which training will be stopped
    mode
        one of {auto, min, max}. In `min` mode,
        training will stop when the quantity
        monitored has stopped decreasing; in `max`
        mode it will stop when the quantity
        monitored has stopped increasing; in `auto`
        mode, the direction is automatically inferred
        from the name of the monitored quantity
    baseline
        Baseline value for the monitored quantity to reach.
        Training will stop if the model doesn't show improvement
        over the baseline
    restore_best_weights
        whether to restore model weights from
        the epoch with the best value of the monitored quantity.
        If False, the model weights obtained at the last step of
        training are used.
    """

    def __init__(
        self,
        parent,
        monitor='val_loss',
        min_delta=0,
        patience=0,
        mode='auto',
        baseline=None,
        restore_best_weights=False,
    ):
        super(EarlyStopping, self).__init__()

        self.parent = parent
        self.monitor = monitor
        self.baseline = baseline
        self.patience = patience
        self.min_delta = min_delta
        self.wait = 0
        self.stopped_epoch = 0
        self.restore_best_weights = restore_best_weights
        self.best_weights = None

        if mode not in ['auto', 'min', 'max']:
            warnings.warn(
                'EarlyStopping mode %s is unknown, ' 'fallback to auto mode.' % mode,
                RuntimeWarning,
            )
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = np.less
        elif mode == 'max':
            self.monitor_op = np.greater
        else:
            if 'acc' in self.monitor:
                self.monitor_op = np.greater
            else:
                self.monitor_op = np.less

        if self.monitor_op == np.greater:
            self.min_delta *= 1
        else:
            self.min_delta *= -1

    def on_train_begin(self, logs=None):
        """on_train_begin part of callback"""
        # Allow instances to be re-used
        self.wait = 0
        self.stopped_epoch = 0
        if self.baseline is not None:
            self.best = self.baseline
        else:
            self.best = np.Inf if self.monitor_op == np.less else -np.Inf

    def on_epoch_end(self, epoch, logs=None):
        """on_epoch_end part of callback"""
        current = self.get_monitor_value(logs)
        if current is None:
            return

        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
            if self.restore_best_weights:
                self.best_weights = self.model.get_weights()
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.model.stop_training = True
                if self.restore_best_weights:
                    lprint('Restoring model weights from the end of ' 'the best epoch')
                    self.model.set_weights(self.best_weights)

        # This is where we stop training externally:
        if self.parent.stop_fitting:
            lprint('Training externally stopped!')
            self.model.stop_training = True

    def on_train_end(self, logs=None):
        """on_train_end part of callback"""
        if self.stopped_epoch > 0:
            lprint('Epoch %05d: early stopping' % (self.stopped_epoch + 1))

    def get_monitor_value(self, logs):
        """returns the monitor value"""
        monitor_value = logs.get(self.monitor)
        if monitor_value is None:
            warnings.warn(
                'Early stopping conditioned on metric `%s` '
                'which is not available. Available metrics are: %s'
                % (self.monitor, ','.join(list(logs.keys()))),
                RuntimeWarning,
            )
        return monitor_value


class ReduceLROnPlateau(Callback):
    """
    Reduce learning rate when a metric has stopped improving.

    Models often benefit from reducing the learning rate by a factor
    of 2-10 once learning stagnates. This callback monitors a
    quantity and if no improvement is seen for a 'patience' number
    of epochs, the learning rate is reduced.

    Example
    -------

    ```python
    reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.2,
                                  patience=5, min_lr=0.001)
    model.fit(X_train, Y_train, callbacks=[reduce_lr])
    ```

    Parameters
    ----------
    monitor
        quantity to be monitored
    factor
        factor by which the learning rate will
        be reduced. new_lr = lr * factor
    patience
        number of epochs with no improvement
        after which learning rate will be reduced
    verbose : int
        0: quiet, 1: update messages
    mode
        one of {auto, min, max}. In `min` mode,
        lr will be reduced when the quantity
        monitored has stopped decreasing; in `max`
        mode it will be reduced when the quantity
        monitored has stopped increasing; in `auto`
        mode, the direction is automatically inferred
        from the name of the monitored quantity
    min_delta
        threshold for measuring the new optimum,
        to only focus on significant changes
    cooldown
        number of epochs to wait before resuming
        normal operation after lr has been reduced
    min_lr
        lower bound on the learning rate
    kwargs
    """

    def __init__(
        self,
        monitor='val_loss',
        factor=0.1,
        patience=10,
        verbose=0,
        mode='auto',
        min_delta=1e-4,
        cooldown=0,
        min_lr=0,
        **kwargs,
    ):
        super(ReduceLROnPlateau, self).__init__()

        self.monitor = monitor
        if factor >= 1.0:
            raise ValueError('ReduceLROnPlateau ' 'does not support a factor >= 1.0.')
        if 'epsilon' in kwargs:
            min_delta = kwargs.pop('epsilon')
            warnings.warn(
                '`epsilon` argument is deprecated and '
                'will be removed, use `min_delta` instead.'
            )
        self.factor = factor
        self.min_lr = min_lr
        self.min_delta = min_delta
        self.patience = patience
        self.verbose = verbose
        self.cooldown = cooldown
        self.cooldown_counter = 0  # Cooldown counter.
        self.wait = 0
        self.best = 0
        self.mode = mode
        self.monitor_op = None
        self._reset()

    def _reset(self):
        """Resets wait counter and cooldown counter."""
        if self.mode not in ['auto', 'min', 'max']:
            warnings.warn(
                'Learning Rate Plateau Reducing mode %s is unknown, '
                'fallback to auto mode.' % (self.mode),
                RuntimeWarning,
            )
            self.mode = 'auto'
        if self.mode == 'min' or (self.mode == 'auto' and 'acc' not in self.monitor):
            self.monitor_op = lambda a, b: np.less(a, b - self.min_delta)
            self.best = np.Inf
        else:
            self.monitor_op = lambda a, b: np.greater(a, b + self.min_delta)
            self.best = -np.Inf
        self.cooldown_counter = 0
        self.wait = 0

    def on_train_begin(self, logs=None):
        self._reset()

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        logs['lr'] = backend.get_value(self.model.optimizer.lr)
        current = logs.get(self.monitor)
        if current is None:
            warnings.warn(
                'Reduce LR on plateau conditioned on metric `%s` '
                'which is not available. Available metrics are: %s'
                % (self.monitor, ','.join(list(logs.keys()))),
                RuntimeWarning,
            )

        else:
            if self.in_cooldown():
                self.cooldown_counter -= 1
                self.wait = 0

            if self.monitor_op(current, self.best):
                self.best = current
                self.wait = 0
            elif not self.in_cooldown():
                self.wait += 1
                if self.wait >= self.patience:
                    old_lr = float(backend.get_value(self.model.optimizer.lr))
                    if old_lr > self.min_lr:
                        new_lr = old_lr * self.factor
                        new_lr = max(new_lr, self.min_lr)
                        backend.set_value(self.model.optimizer.lr, new_lr)
                        if self.verbose > 0:
                            lprint(
                                'Epoch %05d: ReduceLROnPlateau reducing '
                                'learning rate to %s.' % (epoch + 1, new_lr)
                            )
                        self.cooldown_counter = self.cooldown
                        self.wait = 0

    def in_cooldown(self):
        return self.cooldown_counter > 0


class ModelCheckpoint(Callback):
    """
    Save the model after every epoch.

    `filepath` can contain named formatting options,
    which will be filled the value of `epoch` and
    keys in `logs` (passed in `on_epoch_end`).

    For example: if `filepath` is `weights.{epoch:02d}-{val_loss:.2f}.hdf5`,
    then the model checkpoints will be saved with the epoch number and
    the validation loss in the filename.

    Parameters
    ----------
    filepath : string
        path to save the model file.
    monitor
        quantity to monitor
    verbose
        verbosity mode, 0 or 1.
    save_best_only
        if `save_best_only=True`,
        the latest best model according to
        the quantity monitored will not be overwritten.
    save_weights_only
        if True, then only the model's weights will be
        saved (`model.save_weights(filepath)`), else the full model
        is saved (`model.save(filepath)`).
    mode
        one of {auto, min, max}.
        If `save_best_only=True`, the decision
        to overwrite the current save file is made
        based on either the maximization or the
        minimization of the monitored quantity. For `val_acc`,
        this should be `max`, for `val_loss` this should
        be `min`, etc. In `auto` mode, the direction is
        automatically inferred from the name of the monitored quantity.
    period
        Interval (number of epochs) between checkpoints.
    """

    def __init__(
        self,
        filepath,
        monitor='val_loss',
        verbose=0,
        save_best_only=False,
        save_weights_only=False,
        mode='auto',
        period=1,
    ):
        super(ModelCheckpoint, self).__init__()
        self.monitor = monitor
        self.verbose = verbose
        self.filepath = filepath
        self.save_best_only = save_best_only
        self.save_weights_only = save_weights_only
        self.period = period
        self.epochs_since_last_save = 0

        if mode not in ['auto', 'min', 'max']:
            warnings.warn(
                'ModelCheckpoint mode %s is unknown, '
                'fallback to auto mode.' % (mode),
                RuntimeWarning,
            )
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = np.less
            self.best = np.Inf
        elif mode == 'max':
            self.monitor_op = np.greater
            self.best = -np.Inf
        else:
            if 'acc' in self.monitor or self.monitor.startswith('fmeasure'):
                self.monitor_op = np.greater
                self.best = -np.Inf
            else:
                self.monitor_op = np.less
                self.best = np.Inf

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            filepath = self.filepath.format(epoch=epoch + 1, **logs)
            if self.save_best_only:
                current = logs.get(self.monitor)
                if current is None:
                    warnings.warn(
                        'Can save best model only with %s available, '
                        'skipping.' % (self.monitor),
                        RuntimeWarning,
                    )
                else:
                    if self.monitor_op(current, self.best):
                        if self.verbose > 0:
                            lprint(
                                'Epoch %05d: %s=%0.5f improved from %0.5f to %0.5f,'
                                % (epoch + 1, self.monitor, current, self.best, current)
                            )
                        self.best = current
                        if self.save_weights_only:
                            self.model.save_weights(filepath, overwrite=True)
                        else:
                            self.model.save(filepath, overwrite=True)
                    else:
                        if self.verbose > 0:
                            lprint(
                                'Epoch %05d: %s=%0.5f did not improve from %0.5f'
                                % (epoch + 1, self.monitor, current, self.best)
                            )
            else:
                if self.verbose > 0:
                    lprint('Epoch %05d: saving model to %s' % (epoch + 1, filepath))
                if self.save_weights_only:
                    self.model.save_weights(filepath, overwrite=True)
                else:
                    self.model.save(filepath, overwrite=True)


class StopCenterGradient2D(Callback):
    """
    Reset the weights of the center pixels after each training batch.
    """

    def __init__(self, blind_spots=None):
        super(StopCenterGradient2D, self).__init__()
        self.blind_spots = blind_spots

    def on_batch_ends(self, batch, logs=None):
        """
        At the end of each mini-batch, reset the gradient of center pixel to 0.
        """
        b = 0.001
        count = 0
        for lyr in self.model.layers:
            """
            This is to smooth the blind-spot kernel
            """
            if 'dilcv' in lyr.name and '_cv' in lyr.name and count < 3:
                """
                Get weights from the first 3 convolution layers
                """
                # Extract weights values from a layer
                weights = lyr.get_weights()  # (H, W, n_ch_in, n_ch_out or n_filters)

                # Get channel dim size
                num_channels = weights[0].shape[-2]  # n_ch_in

                # Create a smoothing kernel
                kernel = np.array([[b, b, b], [b, 1, b], [b, b, b]])

                if self.blind_spots:
                    for blind_spot in self.blind_spots:
                        spot_after_shift = tuple(map(sum, zip(blind_spot, (1, 1))))
                        kernel[spot_after_shift] = 0

                kernel = kernel[..., np.newaxis, np.newaxis].astype(np.float32)
                kernel /= kernel.sum()

                kernel = np.broadcast_to(
                    kernel, kernel.shape[:2] + (num_channels, num_channels)
                )

                # Apply smoothing kernel to the weights
                weights0 = backend.conv2d(
                    np.transpose(weights[0], [3, 0, 1, 2]),  # (batch, H, W, ch)
                    kernel,  # (H, W, n_ch_in (= n_ch of input tensor), n_ch_out)
                    padding='same',
                )

                # Plug the new weights values back to layer weight array
                weights[0] = np.transpose(weights0, [1, 2, 3, 0])

                # Set layer weights back to layer object
                lyr.set_weights(weights)

                # decrease for upper layers:
                b *= 0.1
                count += 1

        for lyr in self.model.layers:
            """
            This is to reset the gradient of the center of the kernel to 0. Only the first layer.
            """
            if 'dilcv0' in lyr.name and '_cv' in lyr.name:
                # Determine the index of center pixel
                indexes = tuple((i - 1) // 2 for i in lyr.kernel_size)

                # Extract weights values from a layer
                weights = lyr.get_weights()

                # Plug 0 to the center pixel of kernel weight
                weights[0][indexes[0], indexes[1], ...] = 0

                # Set layer weights back to layer object
                lyr.set_weight(weights)

    def on_epoch_end(self, epoch, logs=None):
        """
        This is the same process as self.on_batch_ends.
        To be precise, I'm actually not sure if I need this on epoch again on top of self.on_batch_ends.
        Just to be safe.
        """

        b = 0.001
        count = 0
        for lyr in self.model.layers:
            if 'dilcv' in lyr.name and '_cv' in lyr.name and count < 1:
                weights = lyr.get_weights()  # (H, W, n_ch_in, n_ch_out or n_filters)
                num_channels = weights[0].shape[-2]  # n_ch_in

                kernel = np.array([[b, b, b], [b, 1, b], [b, b, b]])

                if self.blind_spots:
                    for blind_spot in self.blind_spots:
                        spot_after_shift = tuple(map(sum, zip(blind_spot, (1, 1))))
                        kernel[spot_after_shift] = 0

                kernel = kernel[..., np.newaxis, np.newaxis].astype(np.float32)
                kernel /= kernel.sum()

                kernel = np.broadcast_to(
                    kernel, kernel.shape[:2] + (num_channels, num_channels)
                )
                weights0 = backend.conv2d(
                    np.transpose(weights[0], [3, 0, 1, 2]),  # (batch, H, W, ch)
                    kernel,  # (H, W, n_ch_in (= n_ch of input tensor), n_ch_out)
                    padding='same',
                )
                weights[0] = np.transpose(weights0, [1, 2, 3, 0])
                lyr.set_weights(weights)

                # decrease for upper layers:
                b *= 0.1
                count += 1

        for lyr in self.model.layers:
            if 'dilcv0' in lyr.name and '_cv' in lyr.name:
                indexes = tuple((i - 1) // 2 for i in lyr.kernel_size)
                weights = lyr.get_weights()
                weights[0][indexes[0], indexes[1], ...] = 0
                lyr.set_weights(weights)

    def on_train_end(self, logs=None):
        """
        This must be called after training to remove the blind-spot.
        This process is to slightly contaminate some of the original input information to the center pixel.
        So perhaps, a reverse action might be needed on_train_begin if we want to do retrain.
        """
        count = 0
        for lyr in self.model.layers:
            if 'dilcv' in lyr.name and '_cv' in lyr.name and count < 1:
                # Extract weights values from a layer
                weights = lyr.get_weights()  # (H, W, n_ch_in, n_ch_out or n_filters)

                # Get channel dim size
                num_channels = weights[0].shape[-2]  # n_ch_in

                # Create a smoothing kernel
                b = 1
                kernel = np.array([[b, b, b], [b, 1, b], [b, b, b]])

                if self.blind_spots:
                    for blind_spot in self.blind_spots:
                        spot_after_shift = tuple(map(sum, zip(blind_spot, (1, 1))))
                        kernel[spot_after_shift] = 0

                kernel = kernel[..., np.newaxis, np.newaxis].astype(np.float32)
                kernel /= kernel.sum()

                kernel = np.broadcast_to(
                    kernel, kernel.shape[:2] + (num_channels, num_channels)
                )

                # Apply smoothing kernel to the weights
                weights0 = backend.conv2d(
                    np.transpose(weights[0], [3, 0, 1, 2]),  # (batch, H, W, ch)
                    kernel,  # (H, W, n_ch_in (= n_ch of input tensor), n_ch_out)
                    padding='same',
                )

                # Plug the new weights values back to layer weight array
                weights0 = np.transpose(weights0, [1, 2, 3, 0])

                # Determine the index of cneter pixel
                indexes = tuple((i - 1) // 2 for i in lyr.kernel_size)

                # Plug 0 to the center pixel of kernel weight
                original_sum = weights[0].sum()
                a = weights0[indexes[0], indexes[1], ...]
                weights[0][indexes[0], indexes[1], ...] = a
                weights[0] *= original_sum / weights[0].sum()

                # Set layer weights back to layer object
                lyr.set_weights(weights)
                count += 1


class StopCenterGradient3D(Callback):
    """
    Reset the weights of the center pixels after each training batch.
    """

    def __init__(self, blind_spots=None):
        super(StopCenterGradient3D, self).__init__()
        self.blind_spots = blind_spots

    def on_batch_ends(self, batch, logs=None):
        """
        At the end of each mini-batch, reset the gradient of center pixel to 0.
        """
        b = 0.001
        count = 0
        for lyr in self.model.layers:
            """
            This is to smooth the blind-spot kernel
            """
            if 'dilcv' in lyr.name and '_cv' in lyr.name and count < 3:
                """
                Get weights from the first 3 convolution layers
                """
                # Extract weights values from a layer
                weights = lyr.get_weights()  # (H, W, n_ch_in, n_ch_out or n_filters)

                # Get channel dim size
                num_channels = weights[0].shape[-2]  # n_ch_in

                # Create a smoothing kernel
                kernel = np.array(
                    [
                        [[b, b, b], [b, b, b], [b, b, b]],
                        [[b, b, b], [b, 1, b], [b, b, b]],
                        [[b, b, b], [b, b, b], [b, b, b]],
                    ]
                )

                if self.blind_spots:
                    for blind_spot in self.blind_spots:
                        spot_after_shift = tuple(map(sum, zip(blind_spot, (1, 1, 1))))
                        kernel[spot_after_shift] = 0

                kernel = kernel[..., np.newaxis, np.newaxis].astype(np.float32)
                kernel /= kernel.sum()
                kernel = np.broadcast_to(
                    kernel, kernel.shape[:-2] + (num_channels, num_channels)
                )
                weights0 = backend.conv3d(
                    np.transpose(weights[0], [4, 0, 1, 2, 3]),  # (batch, D, H, W, ch)
                    kernel,  # (D, H, W, n_ch_in (= n_ch of input tensor), n_ch_out)
                    padding='same',
                )

                # Plug the new weights values back to layer weight array
                weights[0] = np.transpose(weights0, [1, 2, 3, 4, 0])

                # Set layer weights back to layer object
                lyr.set_weights(weights)

                # decrease for upper layers:
                b *= 0.1
                count += 1

        for lyr in self.model.layers:
            """
            This is to reset the gradient of the center of the kernel to 0. Only the first layer.
            """
            if 'dilcv0' in lyr.name and '_cv' in lyr.name:
                # Determine the index of center pixel
                indexes = tuple((i - 1) // 2 for i in lyr.kernel_size)

                # Extract weights values from a layer
                weights = lyr.get_weights()

                # Plug 0 to the center pixel of kernel weight
                weights[0][indexes[0], indexes[1], indexes[2], ...] = 0

                # Set layer weights back to layer object
                lyr.set_weight(weights)

    def on_epoch_end(self, epoch, logs=None):
        """
        This is the same process as self.on_batch_ends.
        To be precise, I'm actually not sure if I need this on epoch again on top of self.on_batch_ends.
        Just to be safe.
        """
        b = 0.001
        count = 0
        for lyr in self.model.layers:
            if 'dilcv' in lyr.name and '_cv' in lyr.name and count < 1:
                weights = lyr.get_weights()  # (H, W, n_ch_in, n_ch_out or n_filters)
                num_channels = weights[0].shape[-2]  # n_ch_in
                kernel = np.array(
                    [
                        [[b, b, b], [b, b, b], [b, b, b]],
                        [[b, b, b], [b, 1, b], [b, b, b]],
                        [[b, b, b], [b, b, b], [b, b, b]],
                    ]
                )

                if self.blind_spots:
                    for blind_spot in self.blind_spots:
                        spot_after_shift = tuple(map(sum, zip(blind_spot, (1, 1, 1))))
                        kernel[spot_after_shift] = 0

                kernel = kernel[..., np.newaxis, np.newaxis].astype(np.float32)
                kernel /= kernel.sum()
                kernel = np.broadcast_to(
                    kernel, kernel.shape[:-2] + (num_channels, num_channels)
                )
                weights0 = backend.conv3d(
                    np.transpose(weights[0], [4, 0, 1, 2, 3]),  # (batch, D, H, W, ch)
                    kernel,  # (D, H, W, n_ch_in (= n_ch of input tensor), n_ch_out)
                    padding='same',
                )
                weights[0] = np.transpose(weights0, [1, 2, 3, 4, 0])
                lyr.set_weights(weights)

                # decrease for upper layers:
                b *= 0.1
                count += 1

        for lyr in self.model.layers:
            if 'dilcv0' in lyr.name and '_cv' in lyr.name:
                indexes = tuple((i - 1) // 2 for i in lyr.kernel_size)
                weights = lyr.get_weights()
                weights[0][indexes[0], indexes[1], indexes[2], ...] = 0
                lyr.set_weights(weights)

    def on_train_end(self, logs=None):
        """
        This must be called after training to remove the blind-spot.
        This process is to slightly contaminate some of the original input information to the center pixel.
        So perhaps, a reverse action might be needed on_train_begin if we want to do retrain.
        """
        count = 0
        for lyr in self.model.layers:
            if 'dilcv' in lyr.name and '_cv' in lyr.name and count < 1:
                # Extract weights values from a layer
                weights = lyr.get_weights()  # (D, H, W, n_ch_in, n_ch_out or n_filters)

                # Get channel dim size
                num_channels = weights[0].shape[-2]  # n_ch_in

                # Create a smoothing kernel
                b = 1
                # kernel = np.array([[b, b, b], [b, 1, b], [b, b, b]])
                kernel = np.array(
                    [
                        [[b, b, b], [b, b, b], [b, b, b]],
                        [[b, b, b], [b, 1, b], [b, b, b]],
                        [[b, b, b], [b, b, b], [b, b, b]],
                    ]
                )

                if self.blind_spots:
                    for blind_spot in self.blind_spots:
                        spot_after_shift = tuple(map(sum, zip(blind_spot, (1, 1, 1))))
                        kernel[spot_after_shift] = 0

                kernel = kernel[..., np.newaxis, np.newaxis].astype(np.float32)
                kernel /= kernel.sum()
                kernel = np.broadcast_to(
                    kernel, kernel.shape[:-2] + (num_channels, num_channels)
                )

                # Apply smoothing kernel to the weights
                weights0 = backend.conv3d(
                    np.transpose(weights[0], [4, 0, 1, 2, 3]),  # (batch, D, H, W, ch)
                    kernel,  # (D, H, W, n_ch_in (= n_ch of input tensor), n_ch_out)
                    padding='same',
                )

                # Plug the new weights values back to layer weight array
                weights0 = np.transpose(weights0, [1, 2, 3, 4, 0])

                # Determine the index of cneter pixel
                indexes = tuple((i - 1) // 2 for i in lyr.kernel_size)

                # Plug 0 to the center pixel of kernel weight
                original_sum = weights[0].sum()
                a = weights0[indexes[0], indexes[1], indexes[2], ...]
                weights[0][indexes[0], indexes[1], indexes[2], ...] = a
                weights[0] *= original_sum / weights[0].sum()

                # Set layer weights back to layer object
                lyr.set_weights(weights)
                count += 1



def apply_slicing(data, slicing_string: str):
    """
    Function to apply slicing without evaluating argument passed by user

    Parameters
    ----------
    data : numpy.typing.ArrayLike
    slicing_string : str

    Returns
    -------
    sliced_data : numpy.typing.ArrayLike

    """
    if slicing_string == "":
        return data

    try:
        # Get slicing details for each dimension separately
        slices_for_dims = []
        slice_strings = slicing_string[1:-1].split(',')
        for _ in range(len(slice_strings)):  # Iterate over number of dimensions
            slice_string = slice_strings[_]

            # If it is not sliced on a dimension we build slice object to get entire dimension
            if slice_string == ":":
                slices_for_dims.append(slice(data.shape[_]))
            else:
                slice_values = slice_string.split(":")
                slices_for_dims.append(
                    slice(int(slice_values[0]), int(slice_values[1]))
                )

        # Create slice objects for each dimension and form the composite slice object from them
        slicer = [slice(None)] * len(slices_for_dims)
        for axis in range(len(slices_for_dims)):
            slicer[axis] = slices_for_dims[axis]

        # Apply resulting slice object to data array
        sliced_data = data[tuple(slicer)].copy()

    except Exception:
        raise ValueError("Passed slicing argument is not in valid...")

    # Return the sliced array back
    del data
    return sliced_data



from aydin.io.datasets import cropped_newyork

from aydin.it.classic_denoisers.demo.demo_2D_nlm import demo_nlm
from aydin.it.classic_denoisers.nlm import denoise_nlm
from aydin.it.classic_denoisers.test.util_test_nd import check_nd


def test_nlm():
    assert demo_nlm(cropped_newyork(), display=False) >= 0.626 - 0.01


def test_nlm_nd():
    check_nd(denoise_nlm)



# flake8: noqa
import time

import numpy

from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.analysis.image_metrics import mutual_information, spectral_mutual_information
from aydin.io.datasets import normalise, add_blur_2d, dots
from aydin.it.deconvolution.lr_deconv import ImageTranslatorLRDeconv
from aydin.nn.pytorch.it_ptcnn_deconv import PTCNNDeconvolution


def printscore(header, val1, val2, val3, val4):
    print(f"{header}: \t {val1:.4f} \t {val2:.4f} \t {val3:.4f} \t {val4:.4f}")


def demo(image):
    image = normalise(image.astype(numpy.float32))
    blurred_image, psf_kernel = add_blur_2d(image)

    it_deconv = PTCNNDeconvolution(
        max_epochs=1500,
        patience=512,
        learning_rate=0.01,
        normaliser_type='identity',
        psf_kernel=psf_kernel,
    )

    start = time.time()
    it_deconv.train(blurred_image)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    aydin_deconvolved_image = it_deconv.translate(blurred_image)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    lr = ImageTranslatorLRDeconv(
        psf_kernel=psf_kernel, max_num_iterations=20, backend="gputools"
    )
    lr.train(blurred_image)
    lr_deconvolved_image = lr.translate(blurred_image)

    image = numpy.clip(image, 0, 1)
    lr_deconvolved_image = numpy.clip(lr_deconvolved_image, 0, 1)
    aydin_deconvolved_image = numpy.clip(aydin_deconvolved_image, 0, 1)

    print("Below in order: PSNR, norm spectral mutual info, norm mutual info, SSIM: ")
    printscore(
        "n&b image:   ",
        psnr(image, blurred_image),
        spectral_mutual_information(image, blurred_image),
        mutual_information(image, blurred_image),
        ssim(image, blurred_image),
    )

    printscore(
        "lr deconv:    ",
        psnr(image, lr_deconvolved_image),
        spectral_mutual_information(image, lr_deconvolved_image),
        mutual_information(image, lr_deconvolved_image),
        ssim(image, lr_deconvolved_image),
    )

    printscore(
        "aydin deconv: ",
        psnr(image, aydin_deconvolved_image),
        spectral_mutual_information(image, aydin_deconvolved_image),
        mutual_information(image, aydin_deconvolved_image),
        ssim(image, aydin_deconvolved_image),
    )

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(blurred_image, name='blurred')
        viewer.add_image(lr_deconvolved_image, name='lr_deconvolved_image')
        viewer.add_image(aydin_deconvolved_image, name='aydin_deconvolved_image')


image = dots()

demo(image)



from aydin.gui.tabs.qt.base_cropping import BaseCroppingTab


class DenoisingCroppingTab(BaseCroppingTab):
    """
    Cropping Image before denoising

    You may also want to restrict denoising to a region in the image. This is independent from the cropping done for
    training and auto-tuning. Use the sliders to select a region of the image to crop before denoising.
    <moreless>
    <split>
    """



import math

import numpy
from numpy.linalg import norm
from scipy.fft import dctn

from aydin.analysis.resolution_estimate import resolution_estimate


def snr_estimate(image) -> float:
    """Estimates the signal to noise ratio of an image in DB.

    A value of 0 means that the signal and noise have roughly the same energy,
    a negative value means that the noise is stronger than the signal,
    and reciprocally, a positive value means that the signal is stronger
    than the noise.

    Parameters
    ----------
    image : numpy.typing.ArrayLike

    Returns
    -------
    Returns an estimate of the image's signal-to-noise ratio in dB.
    """

    # First we estimate resolution:
    frequency, image = resolution_estimate(image)

    # Compute the DCT:
    image_dct = dctn(image, workers=-1)

    # Compute frequency map:
    f = numpy.zeros_like(image)
    axis_grid = tuple(numpy.linspace(0, 1, s) for s in image.shape)
    for x in numpy.meshgrid(*axis_grid, indexing='ij'):
        f += x ** 2
    f = numpy.sqrt(f)

    # define two domains:
    signal_domain = f <= frequency
    noise_domain = f > frequency

    # First we measure the energy of both signa and noise:
    signal_energy = norm(image_dct[signal_domain]) ** 2
    noise_energy = norm(image_dct[noise_domain]) ** 2

    # However, this is an underestimate of the noise, because we assume that
    # the noise is uniformly distributed in frequency space. Therefore, we need
    # to correct this first estimate. For this we need the 'volume, in frequency
    # space of both domains:

    signal_domain_volume = numpy.sum(signal_domain) / image_dct.size
    noise_domain_volume = 1 - signal_domain_volume

    # We re-estimate the noise energy assuming the same density over the whole
    # spectrum:
    corrected_noise_energy = noise_energy / noise_domain_volume

    # For the signal energy we remove the energy that comes from the noise:
    corrected_signal_energy = (
        signal_energy - signal_domain_volume * corrected_noise_energy
    )

    # We can't let the signal energy to go below zero:
    corrected_signal_energy = max(1e-16, corrected_signal_energy)

    # Signal to noise ratio in dB:
    noise_ratio = 10 * math.log10(corrected_signal_energy / corrected_noise_energy)

    return noise_ratio



import torch
import torch.nn as nn
import torch.nn.functional as F

from aydin.nn.pytorch.models.convblock import ConvBlock


class BabyUnet(nn.Module):
    def __init__(self, n_channel_in=1, n_channel_out=1):
        super(BabyUnet, self).__init__()
        self.pool1 = nn.AvgPool2d(kernel_size=2)
        self.pool2 = nn.AvgPool2d(kernel_size=2)

        self.up1 = lambda x: F.interpolate(x, mode='bilinear', scale_factor=2)
        self.up2 = lambda x: F.interpolate(x, mode='bilinear', scale_factor=2)

        self.conv1 = ConvBlock(n_channel_in, 16)
        self.conv2 = ConvBlock(16, 32)

        self.conv3 = ConvBlock(32, 32)

        self.conv4 = ConvBlock(64, 32)
        self.conv5 = ConvBlock(48, 16)

        self.conv6 = nn.Conv2d(16, n_channel_out, 1)

    def forward(self, x):
        c1 = self.conv1(x)
        x = self.pool1(c1)
        c2 = self.conv2(x)
        x = self.pool2(c2)
        self.conv3(x)

        x = self.up1(x)
        x = torch.cat([x, c2], 1)
        x = self.conv4(x)
        x = self.up2(x)
        x = torch.cat([x, c1], 1)
        x = self.conv5(x)
        x = self.conv6(x)
        return x



import numpy
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, camera, add_noise
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.periodic import PeriodicNoiseSuppressionTransform
from aydin.it.transforms.range import RangeTransform
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import Log


def demo_high_pass_real():

    Log.override_test_exclusion = True
    Log.enable_output = True

    image = normalise(camera().astype(numpy.float32))

    noisy = image.copy()

    freq = 96
    periodic_pattern = 0.1 * (
        1 + numpy.cos(numpy.linspace(0, freq * 2 * numpy.pi, num=image.shape[0]))
    )
    periodic_pattern = periodic_pattern[:, numpy.newaxis]
    noisy += periodic_pattern

    noisy = add_noise(noisy)

    pns = PeriodicNoiseSuppressionTransform()

    preprocessed = pns.preprocess(noisy)
    postprocessed = pns.postprocess(preprocessed)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(preprocessed, name='preprocessed')
        viewer.add_image(postprocessed, name='postprocessed')

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
    )
    regressor = LGBMRegressor(
        patience=128, compute_training_loss=True, max_num_estimators=2048
    )
    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)
    it.add_transform(RangeTransform())

    it.train(noisy)
    denoised_without_preprocessing = it.translate(noisy)

    it.add_transform(pns)
    it.train(noisy)
    denoised_with_preprocessing = it.translate(noisy)

    image = numpy.clip(image, 0, 1)
    denoised_without_preprocessing = numpy.clip(denoised_without_preprocessing, 0, 1)
    denoised_with_preprocessing = numpy.clip(denoised_with_preprocessing, 0, 1)
    psnr_denoised_without = psnr(image, denoised_without_preprocessing)
    ssim_denoised_without = ssim(image, denoised_without_preprocessing)
    psnr_denoised_with = psnr(image, denoised_with_preprocessing)
    ssim_denoised_with = ssim(image, denoised_with_preprocessing)
    print("denoised_without:", psnr_denoised_without, ssim_denoised_without)
    print("denoised_with   :", psnr_denoised_with, ssim_denoised_with)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(preprocessed, name='preprocessed')
        viewer.add_image(postprocessed, name='postprocessed')
        viewer.add_image(
            denoised_without_preprocessing, name='denoised_without_preprocessing'
        )
        viewer.add_image(
            denoised_with_preprocessing, name='denoised_with_preprocessing'
        )


demo_high_pass_real()



# flake8: noqa

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import examples_single
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.attenuation import AttenuationTransform
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log


def demo(image):
    """
    In some cases it might be usefull to append a compression transform (sqrt) after normalisation,
    something akin to a VST transform but without the exact variance stabilisation, and more as a way
    to deskew the histogram. There are only a few situations where this truly helps, and there are not many.
    So by default this is off.
    """

    Log.enable_output = True
    # Log.set_log_max_depth(5)

    generator = StandardFeatureGenerator(
        # include_scale_one=True,
        # include_fine_features=True,
        # include_corner_features=True,
        # include_line_features=True,
        # decimate_large_scale_features=False,
        # extend_large_scale_features=True,
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        # include_spatial_features=True,
    )
    regressor = CBRegressor(patience=20, gpu=True)

    it = ImageTranslatorFGR(
        feature_generator=generator, regressor=regressor, normaliser_transform='sqrt'
    )

    it.train(image, image)
    denoised = it.translate(image)

    ac = AttenuationTransform(axes=0)
    corrected = ac.preprocess(image)

    it.train(corrected, corrected)
    denoised_corrected = it.translate(corrected)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(corrected, name='noisy')
        viewer.add_image(denoised, name='denoised')
        viewer.add_image(denoised_corrected, name='denoised_corrected')


islet = examples_single.royerlab_hcr.get_array().squeeze()
islet = islet[2, :20, 400 : 400 + 256, 700 : 700 + 256]
demo(islet)



import math

import numpy
import pytest

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, camera
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.attenuation import AttenuationTransform
from aydin.it.transforms.salt_pepper import SaltPepperTransform
from aydin.it.transforms.deskew import DeskewTransform
from aydin.it.transforms.fixedpattern import FixedPatternTransform
from aydin.it.transforms.histogram import HistogramEqualisationTransform
from aydin.it.transforms.motion import MotionStabilisationTransform
from aydin.it.transforms.padding import PaddingTransform
from aydin.it.transforms.range import RangeTransform
from aydin.it.transforms.variance_stabilisation import VarianceStabilisationTransform
from aydin.regression.cb import CBRegressor


def test_get_tilling_strategy_and_margins():
    regressor = CBRegressor()
    generator = StandardFeatureGenerator()

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    image = normalise(camera().astype(numpy.float32))[numpy.newaxis, numpy.newaxis, ...]

    tilling_strategy, margins = it._get_tilling_strategy_and_margins(
        image, max_voxels_per_tile=128 * 128, min_margin=8, max_margin=math.inf
    )

    assert tilling_strategy == (1, 1, 4, 4)
    assert margins == (0, 0, 8, 8)

    it = ImageTranslatorFGR(
        feature_generator=generator,
        regressor=regressor,
        max_memory_usage_ratio=0.0000001,
    )

    image = normalise(camera().astype(numpy.float32))[numpy.newaxis, numpy.newaxis, ...]

    tilling_strategy, margins = it._get_tilling_strategy_and_margins(
        image, max_voxels_per_tile=128 * 128, min_margin=8, max_margin=math.inf
    )

    assert tilling_strategy[2] > 4 and tilling_strategy[3] > 4
    assert margins == (0, 0, 8, 8)


def test_stop_training_classic():
    # TODO: implement test
    pass


def test_stop_training_cnn():
    # TODO: implement test
    pass


def test_train():
    # TODO: write a test that checks edge cases for train method inputs
    pass


def test_parse_axes_args_with_bool_lists():
    it = ImageTranslatorFGR()
    sample_batch_arg = [False, True, False, False]
    sample_chan_arg = [True, False, False, False]

    result_batch_arg, result_chan_arg = it.parse_axes_args(
        sample_batch_arg, sample_chan_arg, 4
    )
    assert result_batch_arg == sample_batch_arg
    assert result_chan_arg == sample_chan_arg


def test_parse_axes_args_with_index_lists():
    it = ImageTranslatorFGR()
    sample_batch_arg = [1]
    sample_chan_arg = [0]

    result_batch_arg, result_chan_arg = it.parse_axes_args(
        sample_batch_arg, sample_chan_arg, 4
    )
    assert result_batch_arg == [False, True, False, False]
    assert result_chan_arg == [True, False, False, False]


def test_parse_axes_args_with_bool_lists_3d_spacetime():
    it = ImageTranslatorFGR()
    sample_batch_arg = [False, False, False, False]
    sample_chan_arg = [True, False, False, False]

    result_batch_arg, result_chan_arg = it.parse_axes_args(
        sample_batch_arg, sample_chan_arg, 4
    )
    assert result_batch_arg == sample_batch_arg
    assert result_chan_arg == sample_chan_arg


def test_parse_axes_args_with_index_lists_3d_spacetime():
    it = ImageTranslatorFGR()
    sample_batch_arg = []
    sample_chan_arg = [0]

    result_batch_arg, result_chan_arg = it.parse_axes_args(
        sample_batch_arg, sample_chan_arg, 4
    )
    assert result_batch_arg == [False, False, False, False]
    assert result_chan_arg == [True, False, False, False]


def test_parse_axes_args_with_bool_lists_4d_spacetime():
    it = ImageTranslatorFGR()
    sample_batch_arg = [False, False, False, False, False]
    sample_chan_arg = [True, False, False, False, False]

    result_batch_arg, result_chan_arg = it.parse_axes_args(
        sample_batch_arg, sample_chan_arg, 5
    )
    assert result_batch_arg == sample_batch_arg
    assert result_chan_arg == sample_chan_arg


def test_parse_axes_args_with_index_lists_4d_spacetime():
    it = ImageTranslatorFGR()
    sample_batch_arg = []
    sample_chan_arg = [0]

    result_batch_arg, result_chan_arg = it.parse_axes_args(
        sample_batch_arg, sample_chan_arg, 5
    )
    assert result_batch_arg == [False, False, False, False, False]
    assert result_chan_arg == [True, False, False, False, False]


def test_parse_axes_args_with_all_batch():
    it = ImageTranslatorFGR()

    with pytest.raises(Exception):
        sample_batch_arg = [True, True, True, True]
        sample_chan_arg = [False, False, False, False]

        it.parse_axes_args(sample_batch_arg, sample_chan_arg, 4)

    with pytest.raises(Exception):
        sample_batch_arg = [0, 1, 2, 3]
        sample_chan_arg = []

        it.parse_axes_args(sample_batch_arg, sample_chan_arg, 4)


def test_parse_axes_args_with_all_chan():
    it = ImageTranslatorFGR()

    with pytest.raises(Exception):
        sample_batch_arg = [False, False, False, False]
        sample_chan_arg = [True, True, True, True]

        it.parse_axes_args(sample_batch_arg, sample_chan_arg, 4)

    with pytest.raises(Exception):
        sample_batch_arg = []
        sample_chan_arg = [0, 1, 2, 3]

        it.parse_axes_args(sample_batch_arg, sample_chan_arg, 4)


def test_parse_axes_args_with_colliding_batch_chan():
    it = ImageTranslatorFGR()

    with pytest.raises(Exception):
        sample_batch_arg = [True, False, False, False]
        sample_chan_arg = [True, False, False, False]

        it.parse_axes_args(sample_batch_arg, sample_chan_arg, 4)

    with pytest.raises(Exception):
        sample_batch_arg = [0]
        sample_chan_arg = [0]

        it.parse_axes_args(sample_batch_arg, sample_chan_arg, 4)


def test_parse_axes_args_with_5d_spacetime():
    it = ImageTranslatorFGR()

    with pytest.raises(Exception):
        sample_batch_arg = [False, False, False, False, False]
        sample_chan_arg = [False, False, False, False, False]

        it.parse_axes_args(sample_batch_arg, sample_chan_arg, 5)

    with pytest.raises(Exception):
        sample_batch_arg = []
        sample_chan_arg = []

        it.parse_axes_args(sample_batch_arg, sample_chan_arg, 5)


def test_transform_sorting():
    it = ImageTranslatorFGR()

    it.add_transform(AttenuationTransform())
    it.add_transform(RangeTransform())
    it.add_transform(SaltPepperTransform())
    it.add_transform(DeskewTransform())
    it.add_transform(FixedPatternTransform())
    it.add_transform(HistogramEqualisationTransform())
    it.add_transform(MotionStabilisationTransform())
    it.add_transform(PaddingTransform())
    it.add_transform(VarianceStabilisationTransform())

    for i, t1 in enumerate(it.transforms_list):
        for j, t2 in enumerate(it.transforms_list):
            if i < j:
                assert t1.priority < t2.priority



# -*- coding: utf-8 -*-
#
# aydin documentation build configuration file, created by
# sphinx-quickstart on Sun Jan  4 08:58:22 2015.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
import sys


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath("../../"))

source_dir = os.path.dirname(__file__)
sys.path.insert(0, os.path.join(source_dir, "ext"))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.viewcode",
    "numpydoc",
    "sphinx_rtd_theme",
    # "sphinx_click.ext",
]

numpydoc_show_class_members = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Aydin"
copyright = "2021. Chan Zuckerberg Biohub. All rights reserved"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "default"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {"logo_only": True}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "resources/aydin_logo_grad_black_tr1.png"


# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "aydindoc"



import os
from collections import Counter
from copy import deepcopy
from os.path import exists
from pathlib import Path
from typing import List, Tuple
import numpy
import zarr
import dask.array as da

from aydin.util.log.log import lprint


def is_zarr_storage(input_path):
    """Method to check if given file is a zarr storage or not.

    Parameters
    ----------
    input_path : str

    Returns
    -------
    bool
        Result of whether the file in the given path is a zarr storage or not.

    """
    try:
        z = zarr.open(input_path)
        if len(z.shape) >= 0:
            lprint(f"This path is a ZARR storage: {input_path}")
        else:
            raise Exception
        # IF we reach this point, then we could open the file and therefore it is a Zarr file...
        return True
    except Exception:
        return False


def read_zarr_array(input_path):
    """Method that reads a zarr file. If the file is a zarr.Array, this will
    read as an zarr.Array. If the file is a zarr.Group, this method can only
    read it if it is in dexp-zarr group format.

    Parameters
    ----------
    input_path : str

    Returns
    -------
    numpy.typing.ArrayLike

    """
    g = zarr.open(input_path, mode='a')
    if isinstance(g, zarr.Array):
        return da.from_zarr(input_path)
    else:
        arrays = [
            g[key][key] for key in g.group_keys() if "mip" not in g[key][key].name
        ]

        array = da.stack(arrays, axis=0)

        return array


def get_files_with_most_frequent_extension(path) -> List[str]:
    """Method that looks into the given path and return the list of files with
    the most frequent file extension.

    Parameters
    ----------
    path : str

    Returns
    -------
    List[str]

    """

    files_in_folder = os.listdir(path)

    extensions = [Path(file).suffix[1:] for file in files_in_folder]

    counts = Counter(extensions)

    most_frequent_extension = sorted(counts, key=counts.__getitem__)[-1]

    files = [
        file for file in files_in_folder if file.endswith(f".{most_frequent_extension}")
    ]

    return files


def get_output_image_path(
    path: str, operation_type: str = "denoised"
) -> Tuple[str, int]:
    """Method to get correct output path for given input path and operation type.

    Parameters
    ----------
    path : str
    operation_type : str
        Currently supported values: 'denoised', 'deconvolved', 'hyperstacked'.

    Returns
    -------
    Tuple
        (Correct output path, counter).

    """

    image_formats = [
        ".zarr.zip",
        ".zarr",
        ".tiff",
        ".png",
        ".tif",
        ".TIF",
        ".czi",
        ".npy",
        ".nd2",
    ]

    for image_format in image_formats:
        if image_format in path:
            output_path = (
                f"{path.split(image_format)[0]}_{operation_type}{image_format}"
            )
            output_image_format = image_format
            break
    else:  # means no break in this context
        lprint("Image file format is not supported, will be writing result as tif")
        output_path = f"{path[:path.rfind('.')]}_{operation_type}.tif"
        output_image_format = ".tif"

    counter = 1
    response_counter = None
    while exists(output_path):
        response_counter = counter
        output_path = f"{output_path.split('_denoised')[0]}_{operation_type}{counter}{output_image_format}"
        counter += 1

    return output_path, response_counter


def get_options_json_path(path: str, passed_counter: int = None) -> str:
    options_path = f"{path[:path.rfind('.')]}_options.json"

    if passed_counter is None:
        counter = 1
        while exists(options_path):
            options_path = f"{options_path.split('_options')[0]}_options{counter}.json"
            counter += 1
    else:
        options_path = (
            f"{options_path.split('_options')[0]}_options{passed_counter}.json"
        )

    return options_path


def get_save_model_path(path: str, passed_counter: int = None) -> str:
    model_path = f"{path[:path.rfind('.')]}_model"

    if passed_counter is None:
        counter = 1
        while exists(model_path):
            model_path = f"{model_path.split('_model')[0]}_model{counter}"
            counter += 1
    else:
        model_path = f"{model_path.split('_model')[0]}_model{passed_counter}"

    return model_path


def split_image_channels(image_array, metadata):
    """Method that takes a multichannel image and its metadata and splits
    into single channel images.

    Parameters
    ----------
    image_array : numpy.typing.ArrayLike
    metadata : FileMetadata

    Returns
    -------
    tuple(List[numpy.typing.ArrayLike], List[FileMetadata])
        Tuple of splitted_arrays and metadatas.

    """
    channel_axis = metadata.axes.find("C")

    if channel_axis == -1:
        lprint("Array has no channel axis detected")
        return

    # Handle image splitting
    splitted_arrays = numpy.split(
        image_array, metadata.shape[channel_axis], axis=channel_axis
    )
    splitted_arrays = [numpy.squeeze(array) for array in splitted_arrays]

    # Handle metadata changes
    metadata.batch_axes = tuple(
        x for ind, x in enumerate(metadata.batch_axes) if ind != channel_axis
    )
    metadata.channel_axes = tuple(
        x for ind, x in enumerate(metadata.channel_axes) if ind != channel_axis
    )
    metadata.axes = metadata.axes.replace("C", "")
    metadata.shape = tuple(
        x for idx, x in enumerate(metadata.shape) if idx != channel_axis
    )
    metadata.splitted = True

    metadatas = [metadata] * len(splitted_arrays)

    return splitted_arrays, metadatas


def hyperstack_arrays(image_arrays, metadatas):
    """Method that takes a list of arrays of same shape and their corresponding
    metadatas, then hyperstacks those into a single image.

    Parameters
    ----------
    image_arrays : List[numpy.typing.ArrayLike]
    metadatas : List[FileMetadata]

    Returns
    -------
    tuple(numpy.typing.ArrayLike, FileMetadata)
        Tuple of hyperstacked image array and its metadata.

    """
    if len(image_arrays) < 2:
        lprint("Need at least two images to hyperstack.")
        return image_arrays, metadatas

    shape_of_first_image = ()

    for idx, metadata in enumerate(metadatas):
        if idx == 0:
            shape_of_first_image = metadata.shape
        elif shape_of_first_image != metadata.shape:
            raise Exception(
                "Images are not same shape, hence cannot hyperstack images."
            )

    metadata = deepcopy(metadatas[-1])

    metadata.axes = "B" + metadata.axes
    metadata.batch_axes = (True,) + metadata.batch_axes
    metadata.channel_axes = (False,) + metadata.channel_axes
    metadata.shape = (len(image_arrays),) + metadata.shape
    image_array = numpy.stack(image_arrays)

    return image_array, metadata



import numpy
from scipy.ndimage import shift
from skimage.exposure import rescale_intensity

from aydin.features.groups.translations import TranslationFeatures
from aydin.io.datasets import camera


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def test_translation_feature_group():
    # get image:
    image = n(camera().astype(numpy.float32))

    # Instantiates translation features:
    vectors = [(0, -1), (0, 1), (-1, 0), (1, 0), (-1, -1), (-1, 1), (1, -1), (1, 1)]
    translations = TranslationFeatures(translations=vectors)
    assert translations.num_features(image.ndim) == 8

    # Check receptive field radius:
    assert translations.receptive_field_radius == 1

    # Set image:
    translations.prepare(image)

    # compute features and check their valididty:
    feature = numpy.empty_like(image)
    for index in range(translations.num_features(image.ndim)):
        translations.compute_feature(index=index, feature=feature)
        vector = vectors[index]
        translated = shift(
            image,
            shift=list(vector),
            output=feature,
            order=0,
            mode='constant',
            cval=0.0,
            prefilter=False,
        )
        assert (feature == translated).all()



from aydin.util.log.log import lprint


class CatBoostStopTrainingCallback:
    def __init__(self):
        self.continue_training = True

    def after_iteration(self, info):
        lprint(
            f"{info.iteration}: learn: {info.metrics['learn']['MAE'][-1]}  test: {info.metrics['learn']['MAE'][-1]}"
        )
        return self.continue_training



from typing import Sequence, Optional, Tuple

import numpy
from numpy import ndarray
from scipy.ndimage import convolve, gaussian_filter

from aydin.features.groups.base import FeatureGroupBase
from aydin.util.log.log import lprint


class ConvolutionalFeatures(FeatureGroupBase):
    """
    Convolutional Feature Group class
    """

    def __init__(self, kernels: Optional[Sequence[ndarray]]):
        super().__init__()
        self.kernels = kernels if kernels is None else list(kernels)
        self.image = None
        self.excluded_voxels: Sequence[Tuple[int, ...]] = []

        self.kwargs = None

    @property
    def receptive_field_radius(self) -> int:
        radius = max(max(s // 2 for s in k.shape) for k in self.kernels)
        return radius

    def num_features(self, ndim: int) -> int:
        return len(self.kernels)

    def prepare(self, image, excluded_voxels=None, **kwargs):
        if excluded_voxels is None:
            excluded_voxels = []

        self.image = image
        self.excluded_voxels = list(excluded_voxels)
        self.kwargs = kwargs

    def compute_feature(self, index: int, feature):
        kernel = self.kernels[index]
        lprint(
            f"Convolutional feature: {index} of shape={kernel.shape}, excluded_voxels={self.excluded_voxels}"
        )

        if len(self.excluded_voxels) > 0:
            # No need to do crazy stuff if there is no excluded voxels to start with...

            # We keep the weights that we zero out:
            center_weights = []

            # We keep the slice object for accessing the excluded voxels in the kernel:
            slices = []

            # We need to modify the kernel to take into account the excluded voxels:
            for excluded_voxel in self.excluded_voxels:

                # First we check if the excluded voxel falls within the footprint of the feature:
                if all(
                    (
                        -s // 2 <= v <= s // 2
                        for s, v in zip(kernel.shape, excluded_voxel)
                    )
                ):
                    # Here is the coordinate of the excluded voxel
                    coord = tuple(
                        (s // 2 + v for s, v in zip(kernel.shape, excluded_voxel))
                    )

                    # slice to address the kernel array:
                    aslice = tuple((slice(c, c + 1, None) for c in coord))
                    slices.append(aslice)
                    center_weights.append(float(kernel[aslice]))
                    kernel[aslice] = 0

            # We use a holed-gaussian estimate of the missing value,
            # so we do use the correct weight but avoid using the actual center value
            missing = numpy.zeros_like(kernel)
            for aslice, weight in zip(slices, center_weights):
                missing[aslice] = weight

            # We apply a Gaussian filter to find neighbooring voxels from which we can estimate the missing values:
            missing = gaussian_filter(missing, sigma=1)

            # We zero the excluded voxels from it:
            for aslice in slices:
                missing[aslice] = 0

            # We rescale the missing value estimation kernel:
            missing /= missing.sum()

            # We add the missing-value-estimation to the kernel:
            kernel += missing

        # Convolution:
        convolve(self.image, weights=kernel, output=feature)

    def finish(self):
        self.image = None
        self.excluded_voxels = None
        self.kwargs = None
        self.kernels = None



import importlib
import inspect
import os
import shutil

from aydin.it.base import ImageTranslatorBase
from aydin.it.cnn import ImageTranslatorCNN
from aydin.it.transforms.padding import PaddingTransform
from aydin.it.transforms.range import RangeTransform
from aydin.it.transforms.variance_stabilisation import VarianceStabilisationTransform
from aydin.nn import models
from aydin.restoration.denoise.base import DenoiseRestorationBase
from aydin.util.log.log import lsection


class Noise2SelfCNN(DenoiseRestorationBase):
    """Noise2Self image denoising using "Convolutional Neural Networks" (CNN)"""

    def __init__(
        self,
        *,
        variant: str = 'cnn-jinet',
        use_model=None,
        input_model_path=None,
        lower_level_args=None,
        it_transforms=None,
    ):
        """
        Noise2Self image denoising using "Convolutional Neural Networks" (CNN).

        Parameters
        ----------
        variant : str
            Variant of N2S CNN denoising
        use_model : bool
            Flag to choose to train a new model or infer from a
            previously trained model. By default it is None.
        input_model_path : string
            Path to model that is desired to be used for inference.
            By default it is None.
        """
        super().__init__()
        self.lower_level_args = lower_level_args
        self.backend_it, self.backend_or_model = (
            variant.split("-") if variant is not None else ("cnn", "jinet")
        )

        self.input_model_path = input_model_path
        self.use_model_flag = use_model
        self.model_folder_path = None

        self.it = None
        self.it_transforms = (
            [
                {"class": RangeTransform, "kwargs": {}},
                {"class": PaddingTransform, "kwargs": {}},
                {"class": VarianceStabilisationTransform, "kwargs": {}},
            ]
            if it_transforms is None
            else it_transforms
        )

    @property
    def configurable_arguments(self):
        """Returns the configurable arguments that will be exposed
        on GUI and CLI.
        """

        # IT CNN
        it = ImageTranslatorCNN

        fullargspec3 = inspect.getfullargspec(ImageTranslatorCNN.__init__)

        it_args = {
            "arguments": fullargspec3.args[1:],
            "defaults": fullargspec3.defaults,
            "annotations": fullargspec3.annotations,
            "reference_class": it,
        }

        # Model
        model_modules = DenoiseRestorationBase.get_implementations_in_a_module(models)

        arguments = {}

        for module in model_modules:
            model_args = self.get_class_implementation_kwonlyargs(
                models, module, module.name + "Model"
            )
            arguments["Noise2SelfCNN-" + module.name] = {
                "model": model_args,
                "it": it_args,
            }

        return arguments

    @property
    def implementations(self):
        """Returns the list of discovered implementations for given method."""
        return [
            "Noise2SelfCNN-" + x.name
            for x in self.get_implementations_in_a_module(models)
        ]

    @property
    def implementations_description(self):
        cnn_description = Noise2SelfCNN.__doc__.strip()

        descriptions = []

        for module in self.get_implementations_in_a_module(models):
            response = importlib.import_module(models.__name__ + '.' + module.name)
            elem = [
                x for x in dir(response) if module.name.replace("_", "") in x.lower()
            ][
                0
            ]  # class name

            elem_class = response.__getattribute__(elem)
            descriptions.append(
                cnn_description
                + "<br><br>"
                + elem_class.__doc__.replace("\n\n", "<br><br>")
            )

        return descriptions

    def stop_running(self):
        """Method to stop running N2S instance"""
        self.it.stop_training()

    def get_translator(self):
        """Returns the corresponding translator instance for given selections.

        Returns
        -------
        it : ImageTranslatorBase

        """
        # Use a pre-saved model or train a new one from scratch and save it
        if self.use_model_flag:
            # Unarchive the model file and load its ImageTranslator object into self.it
            shutil.unpack_archive(
                self.input_model_path, os.path.dirname(self.input_model_path), "zip"
            )
            it = ImageTranslatorBase.load(self.input_model_path[:-4])
        else:
            it = ImageTranslatorCNN(
                **self.lower_level_args["it"]["kwargs"]
                if self.lower_level_args is not None
                else {}
            )

        return it

    def add_transforms(self):
        if self.it_transforms is not None:
            for transform in self.it_transforms:
                transform_class = transform["class"]
                transform_kwargs = transform["kwargs"]
                self.it.add_transform(transform_class(**transform_kwargs))

    def train(
        self, noisy_image, *, batch_axes=None, chan_axes=None, image_path=None, **kwargs
    ):
        """Method to run Noise2Self CNN training.

        Parameters
        ----------
        noisy_image : numpy.ndarray
        batch_axes : array_like, optional
            Indices of batch axes.
        chan_axes : array_like, optional
            Indices of channel axes.
        image_path : str

        Returns
        -------
        response : numpy.ndarray

        """
        with lsection("Noise2Self train is starting..."):
            if any(chan_axes):
                return

            self.it = self.get_translator()

            self.add_transforms()

            # Train a new model
            self.it.train(
                noisy_image,
                noisy_image,
                batch_axes=batch_axes,
                channel_axes=chan_axes,
                train_valid_ratio=kwargs['train_valid_ratio']
                if 'train_valid_ratio' in kwargs
                else 0.1,
                callback_period=kwargs['callback_period']
                if 'callback_period' in kwargs
                else 3,
                jinv=kwargs['jinv'] if 'jinv' in kwargs else None,
            )

            # Save the trained model
            # self.save_model(image_path)  # TODO:  fix the problems here

    def denoise(self, noisy_image, *, batch_axes=None, chan_axes=None, **kwargs):
        """Method to denoise an image with trained Noise2Self.

        Parameters
        ----------
        batch_axes : array_like, optional
            Indices of batch axes.
        chan_axes : array_like, optional
            Indices of channel axes.
        noisy_image : numpy.ndarray

        Returns
        -------
        response : numpy.ndarray

        """
        with lsection("Noise2Self denoise is starting..."):

            # Predict the resulting image
            response = self.it.translate(
                noisy_image,
                batch_axes=batch_axes,
                channel_axes=chan_axes,
                tile_size=kwargs['tile_size'] if 'tile_size' in kwargs else None,
            )

            response = response.astype(noisy_image.dtype, copy=False)

            return response


def noise2self_cnn(image, *, batch_axes=None, chan_axes=None, variant=None):
    """Method to denoise an image with Noise2Self CNN.

    Parameters
    ----------
    image : numpy.ndarray
        Image to denoise
    batch_axes : array_like, optional
        Indices of batch axes.
    chan_axes : array_like, optional
        Indices of channel axes.
    variant : str
        Algorithm variant.


    Returns
    -------
    Denoised image : numpy.ndarray

    """
    # Run N2S and save the result
    n2s = Noise2SelfCNN(variant=variant)

    # Train
    n2s.train(image, batch_axes=batch_axes, chan_axes=chan_axes)

    # Denoise
    denoised = n2s.denoise(image, batch_axes=batch_axes, chan_axes=chan_axes)

    return denoised



# flake8: noqa
from aydin.io.datasets import cropped_newyork
from aydin.util.j_invariance.demo.demo_j_invariant_classic import (
    demo_j_invariant_classic,
)


def test_j_invariant_classic():
    demo_j_invariant_classic(cropped_newyork(crop_amount=470), display=False)



# flake8: noqa
import numpy
from scipy.ndimage import shift

from aydin.util.fast_shift.fast_shift import fast_shift
from aydin.io.datasets import fibsem, examples_single
from aydin.util.log.log import lsection, Log


def demo_fast_shift(image_name: str, image, _shift, repeats=32):
    Log.enable_output = True

    with lsection(f"shift={_shift}:"):
        with lsection(f"Numba-CPU {image_name} (r={repeats}):"):
            for _ in range(repeats):
                numba_shifted_image = fast_shift(image, shift=_shift)

        with lsection(f"Scipy {image_name} (r={repeats}):"):
            for _ in range(repeats):
                scipy_shifted_image = shift(image, shift=_shift)

    # import napari
    #
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(
    #         numba_shifted_image, name='numba_shifted_image', colormap='plasma'
    #     )
    #     viewer.add_image(
    #         scipy_shifted_image, name='scipy_shifted_image', colormap='plasma'
    #     )
    #     viewer.add_image(
    #         numpy.abs(numba_shifted_image - scipy_shifted_image),
    #         name='numba_shifted_image - scipy_shifted_image',
    #         colormap='plasma',
    #     )
    #
    # numpy.testing.assert_array_almost_equal(
    #     numba_shifted_image, scipy_shifted_image, decimal=3
    # )


image_2d = fibsem(full=True).astype(numpy.float32)
image_3d = examples_single.royerlab_hcr.get_array().squeeze()[2]
image_4d = examples_single.hyman_hela.get_array().squeeze()


for _shift in [(3, -1), (17, -51), (-317, 511)]:
    demo_fast_shift("fibsem 2D", image_2d, _shift=_shift, repeats=1)

for _shift in [(3, -1, 2), (17, -51, 7), (-317, 511, -128)]:
    demo_fast_shift("islet 3D", image_3d, _shift=_shift, repeats=1)

for _shift in [(3, -1, 2, -7), (17, -51, -20, 10), (-317, 511, -126, 212)]:
    demo_fast_shift("hela 4D", image_4d, _shift=_shift, repeats=1)



# flake8: noqa
import time

import numpy
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import normalise, add_noise, newyork
from aydin.it.cnn import ImageTranslatorCNN


def demo(image, max_epochs=10):
    image = normalise(image.astype(numpy.float32))
    noisy = add_noise(image)

    # CNN based Image translation:
    # input_dim only includes H, W, C; number of images is not included
    it = ImageTranslatorCNN(
        training_architecture='checkran',  # 'checkran',  #
        model_architecture='jinet',  # 'jinet',  #
        nb_unet_levels=2,
        batch_norm='instance',  # None,  #
        activation='ReLU',
        # patch_size=512,
        mask_size=3,
        # total_num_patches=400,
        # batch_size=40,
        max_epochs=max_epochs,
        learn_rate=0.01,
        # reduce_lr_factor=0.1,  # 0.3 for masking methods, 0.1 for jinet
        # patience=10,
    )
    it.verbose = 1

    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    # in case of batching we have to do this:
    start = time.time()
    denoised_inf = it.translate(noisy, tile_size=512)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy.reshape(image.shape), 0, 1)
    denoised_inf = numpy.clip(denoised_inf, 0, 1)
    print("noisy       :", psnr(image, noisy), ssim(noisy, image))
    print("denoised_inf:", psnr(image, denoised_inf), ssim(denoised_inf, image))

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(normalise(image), name='image')
        viewer.add_image(normalise(noisy), name='noisy')
        viewer.add_image(normalise(denoised_inf), name='denoised_inf')


# camera_image = camera()
# demo(camera_image)
# lizard_image = lizard()
# demo(lizard_image)
# pollen_image = pollen()
# demo(pollen_image)
newyork_image = newyork()
demo(newyork_image)
# characters_image = characters()
# demo(characters_image)
# fibsem_image = fibsem()
# demo(fibsem_image)



# flake8: noqa
import numpy as np
from aydin.it.cnn.util.mask_generator import masker, maskedgen, randmaskgen


def demo_maskgen():  # TODO: refactor this demo into a test
    batch_vol = (64, 64)
    mask_shape = (3, 3)
    batch_size = 8
    image = np.ones((64, 64, 64, 1))

    masked_chkr = maskedgen(batch_vol, mask_shape, image, batch_size)
    masked_rndm = randmaskgen(image, batch_size, p_maskedpixels=0.1)

    for m_rndm in masked_rndm:
        break
    for m_chkr in masked_chkr:
        break

    a = masker(batch_vol, 0, mask_shape, 0.1)
    b = masker(batch_vol, 0, mask_shape, None)



import numpy


def bbox_idx(x, thresh=None):
    """
    Returns indices of bounding box of receptive filed. This is useful for visualize receptive field.

    Parameters
    ----------
    x
        output image of the target model with inpulse input image
    thresh

    Returns
    -------
    list of indices of bounding box of receptive filed
    """
    if thresh is None:
        thresh = 0
    return [[numpy.amin(i), numpy.amax(i)] for i in numpy.where(x != thresh)]


# Derive receptive field from the model
def receptive_field_conv(kernel, stride, n0=1, n_lyrs=1):
    """
    Compute receptive field for convolution layers

    Parameters
    ----------
    kernel
        kernel size
    stride
        stride size
    n0
        receptive field from previous layer
    n_lyrs
        number of layers

    Returns
    -------
    receptive field
    """
    n1 = n0
    for i in range(n_lyrs):
        n1 = kernel + (n1 - 1) * stride
    return n1


def receptive_field_pool(kernel, n0=1, shift_n=0, n_lyrs=1):
    """
    Compute receptive field for pooling layers

    Parameters
    ----------
    kernel
        kernel size
    n0
        receptive field from previous layer
    shift_n
        number of shifting pixels in shift convolution architecture
    n_lyrs
        number of layers

    Returns
    -------
    a tuple of two n1, s
    """
    n1 = n0 * kernel ** n_lyrs
    if shift_n > 1:  # shift_n is the Nst pooling lyr
        s = 1 + 2 ** shift_n
    else:
        s = shift_n
    return n1, s


def receptive_field_up(pl_size, n0=1):
    """
    Compute receptive field for up sampling layers

    Parameters
    ----------
    pl_size
        pooling size
    n0
        receptive field from previous layer

    Returns
    -------
    receptive field
    """
    return n0 if n0 == 1 else numpy.ceil(n0 * 1 / pl_size).astype(int)


def receptive_field_model(model, verbose=False):
    """
    Compute theoretical receptive field. Effective receptive field could be a square root of it.

    Parameters
    ----------
    model
        CNN model
    verbose

    Returns
    -------
    theoretical receptive field
    """
    n1 = 1  # starting field size to calculate receptive field
    shift_n = 0  # index of pooling lyr to calc. shift in shiftconv
    s = 0  # shift due to shiftconv
    shift = False
    layers = numpy.copy(model.layers)
    layers = list(layers)
    layers.reverse()
    for layer in layers:
        lyr_type = layer.__class__.__name__
        if 'Cropping' in lyr_type:
            shift = True
        elif 'Conv' in lyr_type:
            n1 = receptive_field_conv(layer.kernel_size[0], layer.strides[0], n1)
        elif 'Pooling' in lyr_type:
            if shift:
                shift_n += 1
            n1, s = receptive_field_pool(layer.pool_size[0], n1, shift_n)
        elif 'UpSampling' in lyr_type:
            n1 = receptive_field_up(layer.size[0], n1)
        if verbose:
            print(f'{lyr_type}: RF {n1}, shift {s}')
    if shift:
        n1 = (n1 + s) * 2
        print(
            f'Synthetic receptive field is {n1}^2 (with {int(n1/2 + s)}^2 empty space at 4 corners).'
        )
    return n1



from functools import partial
from typing import Optional

import numpy
from scipy.ndimage import gaussian_filter

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart


def calibrate_denoise_gaussian(
    image,
    max_sigma: float = 2,
    max_num_truncate: int = 4,
    crop_size_in_voxels: Optional[int] = 128000,
    max_num_evaluations: int = 256,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the Gaussian denoiser for the given image and returns the optimal
    parameters obtained using the N2S loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate denoiser for.

    max_sigma: float
        Maximum sigma for Gaussian filter.

    max_num_truncate: int
        Maximum number of Gaussian filter truncations to try.
        (advanced)

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate
        denoiser.
        (advanced)

    max_num_evaluations: int
        Maximum number of evaluations for finding
        the optimal parameters.
        (advanced)

    display_images: bool
        When True the denoised images encountered
        during optimisation are shown

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """
    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # Size range:
    sigma_range = (0.0, max(0.0, max_sigma) + 1e-9)  # numpy.arange(0.2, 2, 0.1) ** 1.5

    # Truncate range (order matters: we want 4 -- the default -- first):
    truncate_range = [4, 8, 2, 1][: min(max_num_truncate, 4)]

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {'sigma': sigma_range, 'truncate': truncate_range}

    # Partial function:
    _denoise_gaussian = partial(denoise_gaussian, **other_fixed_parameters)

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_smart(
            crop,
            _denoise_gaussian,
            denoise_parameters=parameter_ranges,
            display_images=display_images,
            max_num_evaluations=max_num_evaluations,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = 2 * image.nbytes

    return denoise_gaussian, best_parameters, memory_needed


def denoise_gaussian(image, sigma: float = 1, truncate: float = 4, **kwargs):
    """
    Denoises the given image using a simple Gaussian filter.
    Difficult to beat in terms of speed and often provides
    sufficient although not superb denoising performance. You
    should always try simple and fast denoisers first, and see
    if that works for you. If it works and is sufficient for
    your needs, why go for slower and more complex and slower
    approach? The only weakness of gaussian filtering is that it
    affects all frequencies. In contrast, the auto-tuned Butterworth
    denoiser will not blur within the estimated band-pass of
    the signal. Thus we recommend you use the Butterworth denoiser
    instead unless you have a good reason this use this one.
    \n\n
    Note: We recommend applying a variance stabilisation transform
    to improve results for images with non-Gaussian noise.

    Parameters
    ----------
    image: ArrayLike
        nD image to denoise

    sigma: float
        Standard deviation for Gaussian kernel.

    truncate: float
         Truncate the filter at this many standard deviations.

    Returns
    -------
    Denoised image
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    return gaussian_filter(image, sigma=sigma, truncate=truncate)



import math
from functools import partial
from typing import Optional, Union, Tuple, Sequence
import numpy
from numba import jit, prange
from numpy.fft import fftshift, ifftshift
from scipy.fft import fftn, ifftn, dctn, idctn, dstn, idstn

from aydin.util.array.outer import outer_sum
from aydin.util.crop.rep_crop import representative_crop
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart
from aydin.util.patch_size.patch_size import default_patch_size
from aydin.util.patch_transform.patch_transform import (
    extract_patches_nd,
    reconstruct_from_nd_patches,
)


def calibrate_denoise_spectral(
    image,
    axes: Optional[Tuple[int, ...]] = None,
    patch_size: Optional[Union[int, Tuple[int], str]] = None,
    try_dct: bool = True,
    try_fft: bool = True,
    try_dst: bool = False,
    max_order: float = 6.0,
    crop_size_in_voxels: Optional[int] = None,
    max_num_evaluations: int = 256,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the Spectral denoiser for the given image and returns the optimal
    parameters obtained using the N2S loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate spectral denoiser for.

    axes: Optional[Tuple[int,...]]
        Axes over which to apply the spectral transform (dct, dst, fft) for denoising each patch.

    patch_size: int
        Patch size for the 'image-to-patch' transform.
        Can be: 'full' for a single patch covering the whole image, 'half', 'quarter',
        or an int s that corresponds to isotropic patches of shape: (s,)*image.ndim,
        or a tuple of ints. By default (None) the patch size is chosen automatically
        to give the best results.
        (advanced)

    try_dct: bool
        Tries DCT transform during optimisation.

    try_fft: bool
        Tries FFT transform during optimisation.

    try_dst: bool
        Tries DST ransform during optimisation.

    max_order: float
        Maximal order for the Butterworth filter.
        (advanced)

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate denoiser.
        (advanced)

    max_num_evaluations: int
        Maximum number of evaluations for finding the optimal parameters.
        (advanced)

    display_images: bool
        When True the denoised images encountered during optimisation are shown

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """
    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels, display_crop=False)

    # Normalise patch size:
    patch_size = default_patch_size(image, patch_size, odd=True)

    # Ranges:
    threshold_range = (0.0, 1.0)  # np.arange(0, 0.5, 0.02) ** 2
    freq_bias_range = (0.0, 2.0)  # np.arange(0, 2, 0.2)
    freq_cutoff_range = (0.01, 1.0)
    order_range = (0.5, max_order)

    # prepare modes list
    modes = []
    if try_dct:
        modes.append("dct")
    if try_fft:
        modes.append("fft")
    if try_dst:
        modes.append("dst")

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {
        'threshold': threshold_range,
        'freq_bias_stength': freq_bias_range,
        # 'reconstruction_gamma': [0.0001, 0.1, 1.0],
        'freq_cutoff': freq_cutoff_range,
        'order': order_range,
        'mode': modes,
    }  # 'fft',, 'fft'

    # Combine fixed parameters:
    other_fixed_parameters = other_fixed_parameters | {
        'patch_size': patch_size,
        'axes': axes,
    }

    # Partial function:
    _denoise_spectral = partial(
        denoise_spectral, **(other_fixed_parameters | {'multi_core': False})
    )

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_smart(
            crop,
            _denoise_spectral,
            denoise_parameters=parameter_ranges,
            display_images=display_images,
            max_num_evaluations=max_num_evaluations,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = 2 * image.nbytes + 8 * image.nbytes * math.prod(patch_size)

    return denoise_spectral, best_parameters, memory_needed


def denoise_spectral(
    image,
    axes: Optional[Tuple[int, ...]] = None,
    patch_size: Optional[Union[int, Tuple[int], str]] = None,
    mode: str = 'dct',
    threshold: float = 0.5,
    freq_bias_stength: float = 1,
    freq_cutoff: Union[float, Sequence[float]] = 0.5,
    order: float = 1,
    reconstruction_gamma: float = 0,
    multi_core: bool = True,
):
    """Denoises the given image by first applying the patch
    transform, and then zeroing Fourier/DCT/DST coefficients
    below a given threshold. In addition, we apply Butterworth
    filter to suppress frequencies above the band-pass and a
    configurable frequency bias before applying the thresholding
    to favour suppressing high versus low frequencies.
    \n\n
    Note: This seems like a lot of parameters, but thanks to our
    auto-tuning approach these parameters are all automatically
     determined 😊.

    Parameters
    ----------
    image: ArrayLike
        Image to denoise

    axes: Optional[Tuple[int,...]]
        Axes over which to apply the spetcral transform (dct, dst, fft) for denoising each patch.

    patch_size: int
        Patch size for the 'image-to-patch' transform.
        Can be: 'full' for a single patch covering the whole image, 'half', 'quarter',
        or an int s that corresponds to isotropic patches of shape: (s,)*image.ndim,
        or a tuple of ints. By default (None) the patch size is chosen automatically
        to give the best results.

    mode: str
        Possible modes are: 'dct'(works best!), 'dst', and 'fft'.

    threshold: float
        Threshold between 0 and 1

    freq_bias_stength: float
        Frequency bias: closer to zero: no bias against high frequencies,
        closer to one and above: stronger bias towards high-frequencies.

    freq_cutoff: float
        Cutoff frequency, must be within [0, 1]. In addition

    order: float
        Filter order, typically an integer above 1.

    reconstruction_gamma: float
        Patch reconstruction parameter

    multi_core: bool
        By default we use as many cores as possible, in some cases, for small
        (test) images, it might be faster to run on a single core instead of
        starting the whole parallelization machinery.


    Returns
    -------
    Denoised image

    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # 'full' patch size:
    if patch_size == 'full':
        patch_size = image.shape
    elif patch_size == 'half':
        patch_size = tuple(max(3, 2 * (s // 4)) for s in image.shape)
    elif patch_size == 'quarter':
        patch_size = tuple(max(3, 4 * (s // 8)) for s in image.shape)

    # Normalise patch size:
    patch_size = default_patch_size(image, patch_size, odd=True)

    # Default axes:
    if axes is None:
        axes = tuple(range(image.ndim))

    # Selected axes:
    selected_axes = tuple((a in axes) for a in range(image.ndim))

    workers = -1 if multi_core else 1
    axes = tuple(a for a in range(1, image.ndim + 1) if (a - 1) in axes)
    if mode == 'fft':
        transform = lambda x: fftshift(  # noqa: E731
            fftn(x, workers=workers, axes=axes), axes=axes
        )
        i_transform = lambda x: ifftn(  # noqa: E731
            ifftshift(x, axes=axes), workers=workers, axes=axes
        )
    elif mode == 'dct':
        transform = partial(dctn, workers=workers, axes=axes)
        i_transform = partial(idctn, workers=workers, axes=axes)
    elif mode == 'dst':
        transform = partial(dstn, workers=workers, axes=axes)
        i_transform = partial(idstn, workers=workers, axes=axes)
    else:
        raise ValueError(f"Unsupported mode: {mode}")

    # Normalise freq_cutoff argument to tuple:
    if type(freq_cutoff) is not tuple:
        freq_cutoff = tuple((freq_cutoff,) * image.ndim)

    # First we apply the patch transform:
    patches = extract_patches_nd(image, patch_size=patch_size)

    # ### PART 1: apply Butterworth filter to patches:

    # Then we apply the sparsifying transform:
    patches = transform(patches)

    # Compute adequate squared distance image and chose filter implementation:
    if mode == 'fft':
        f = _compute_distance_image_for_fft(freq_cutoff, patch_size, selected_axes)
    elif mode == 'dct' or mode == 'dst':
        f = _compute_distance_image_for_dxt(freq_cutoff, patch_size, selected_axes)
    else:
        raise ValueError(f"Unsupported mode: {mode}")

    # Configure filter function:
    filter_wrapped = jit(nopython=True, parallel=multi_core)(_filter)

    # Apply filter:
    patches = filter_wrapped(patches, f, order)

    # ### PART 2: apply thresholding:

    # Window for frequency bias:
    freq_bias = _freq_bias_window(patch_size, freq_bias_stength)

    # We use this value to estimate power per coefficient:
    power = numpy.absolute(patches)
    power *= freq_bias

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(patches, name='patches')
    #     viewer.add_image(f_patches, name='f_patches')
    #     viewer.add_image(power, name='power')
    #     viewer.add_image(freq_bias, name='freq_bias')

    # What is the max coefficient in the transforms:
    max_value = numpy.max(power)

    # We derive from that the threshold:
    threshold *= max_value

    # Here are the entries that are below the threshold:
    below = power < threshold

    # Thresholding:
    patches[below] = 0

    # Transform back to real space:
    patches = i_transform(patches)

    # convert to real:
    if numpy.iscomplexobj(patches):
        patches = numpy.real(patches)

    # Transform back from patches to image:
    denoised_image = reconstruct_from_nd_patches(
        patches, image.shape, gamma=reconstruction_gamma
    )

    # Cast back to float32 if needed:
    denoised_image = denoised_image.astype(numpy.float32, copy=False)

    return denoised_image


def _freq_bias_window(shape: Tuple[int], alpha: float = 1):
    window_tuple = tuple(numpy.linspace(0, 1, s) ** 2 for s in shape)
    window_nd = numpy.sqrt(outer_sum(*window_tuple)) + 1e-6
    window_nd = 1.0 / (1.0 + window_nd)
    window_nd **= alpha
    window_nd /= window_nd.max()
    window_nd = window_nd.astype(numpy.float32)
    return window_nd


def _compute_distance_image_for_dxt(freq_cutoff, shape, selected_axes):

    # Normalise selected axes:
    if selected_axes is None:
        selected_axes = (a for a in range(len(shape)))

    f = numpy.zeros(shape=shape, dtype=numpy.float32)
    axis_grid = tuple(
        (numpy.linspace(0, 1, s) if sa else numpy.zeros((s,)))
        for sa, s in zip(selected_axes, shape)
    )
    for fc, x in zip(freq_cutoff, numpy.meshgrid(*axis_grid, indexing='ij')):
        f += (x / fc) ** 2
    return f


def _compute_distance_image_for_fft(freq_cutoff, shape, selected_axes):
    f = numpy.zeros(shape=shape, dtype=numpy.float32)
    axis_grid = tuple(
        (numpy.linspace(-1, 1, s) if sa else numpy.zeros((s,)))
        for sa, s in zip(selected_axes, shape)
    )
    for fc, x in zip(freq_cutoff, numpy.meshgrid(*axis_grid, indexing='ij')):
        f += (x / fc) ** 2
    return f


def _filter(image_f, f, order):
    factor = (1.0 + numpy.sqrt(f) ** (2 * order)) ** (-0.5)
    factor = factor.astype(numpy.float32)
    n = image_f.shape[0]
    for i in prange(n):
        image_f[i] *= factor
    return image_f



# flake8: noqa

import numpy
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.analysis.find_kernel import compute_relative_blur_kernel
from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import newyork, pollen, normalise, add_noise, lizard, dots
from aydin.it.deconvolution.lr_deconv import ImageTranslatorLRDeconv
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor


def demo(image, name):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """

    # Log.set_log_max_depth(5)

    image = normalise(image.astype(numpy.float32))
    noisy = add_noise(image)

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=False,
        include_fine_features=True,
        include_line_features=True,
        include_spatial_features=True,
    )
    regressor = CBRegressor(patience=16)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    it.train(image, image)
    denoised_clean = it.translate(image)

    computed_kernel = compute_relative_blur_kernel(denoised_clean, image)

    psf_kernel = numpy.asarray([[0.5, 1, 0.5], [1, 1, 1], [0.5, 1, 0.5]])
    psf_kernel = psf_kernel / psf_kernel.sum()

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(normalise(computed_kernel), name='computed_kernel')
        viewer.add_image(normalise(psf_kernel), name='psf_kernel')

    it.train(noisy, noisy)
    denoised = it.translate(noisy)

    lr = ImageTranslatorLRDeconv(
        psf_kernel=computed_kernel, max_num_iterations=3, backend='cupy'
    )
    lr.train(denoised, denoised)
    denoised_deconv = lr.translate(denoised)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)
    denoised_deconv = numpy.clip(denoised_deconv, 0, 1)

    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)

    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)

    psnr_denoised_deconv = psnr(image, denoised_deconv)
    ssim_denoised_deconv = ssim(image, denoised_deconv)

    print("noisy          :", psnr_noisy, ssim_noisy)
    print("denoised       :", psnr_denoised, ssim_denoised)
    print("denoised_deconv:", psnr_denoised_deconv, ssim_denoised_deconv)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(
            normalise(numpy.log1p(numpy.absolute(computed_kernel))),
            name='computed_kernel',
        )
        viewer.add_image(normalise(image), name='image')
        viewer.add_image(normalise(noisy), name='noisy')
        viewer.add_image(normalise(denoised), name='denoised')
        viewer.add_image(normalise(denoised_deconv), name='denoised_deconv')


newyork_image = newyork()
demo(newyork_image, "newyork")
camera_image = camera()
demo(camera_image, "camera")
lizard_image = lizard()
demo(lizard_image, "lizard")
pollen_image = pollen()
demo(pollen_image, "pollen")
dots_image = dots()
demo(dots_image, "dots")



import time
from os.path import join
import numpy
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, add_noise
from aydin.io.folders import get_temp_folder
from aydin.regression.base import RegressorBase
from aydin.regression.cb import CBRegressor
from aydin.regression.lgbm import LGBMRegressor
from aydin.regression.linear import LinearRegressor
from aydin.regression.nn import NNRegressor
from aydin.regression.rf import RFRegressor
from aydin.regression.sv import SVRegressor


def test_linear_regressor():
    regressor = LinearRegressor()
    with_regressor(regressor, min_ssim=0.6)


def test_rf_regressor():
    regressor = RFRegressor()
    with_regressor(regressor, min_ssim=0.6)


def test_svr_regressor():
    regressor = SVRegressor()
    with_regressor(regressor, min_ssim=0.65)


def test_lgbm_regressor():
    regressor = LGBMRegressor(max_num_estimators=600)
    with_regressor(regressor, min_ssim=0.75)


def test_cb_regressor():
    regressor = CBRegressor(max_num_estimators=600)
    with_regressor(regressor, min_ssim=0.75)


def test_nn_regressor():
    regressor = NNRegressor(max_epochs=6, depth=6)
    with_regressor(regressor, min_ssim=0.64)


def with_regressor(regressor, min_ssim=0.8):

    image = camera().astype(numpy.float32)
    image = normalise(image)
    noisy = add_noise(image)

    # feature generator requires images in 'standard' form: BCTZYX
    noisy = noisy[numpy.newaxis, numpy.newaxis, ...]

    generator = StandardFeatureGenerator(max_level=6)

    features = generator.compute(noisy, exclude_center_value=True)

    x = features.reshape(-1, features.shape[-1])
    y = noisy.reshape(-1)

    regressor.fit(x, y)

    yp = regressor.predict(x)

    denoised = yp.reshape(image.shape)

    denoised = numpy.clip(denoised, 0, 1)

    ssim_value = ssim(denoised, image)
    psnr_value = psnr(denoised, image)

    print("denoised", psnr_value, ssim_value)

    assert ssim_value > min_ssim

    # Test saveload
    temp_file = join(get_temp_folder(), "test_reg_saveload.json" + str(time.time()))
    regressor.save(temp_file)

    del regressor

    loaded_regressor = RegressorBase.load(temp_file)

    yp = loaded_regressor.predict(x)

    denoised = yp.reshape(image.shape)

    # print(numpy.min(denoised), numpy.min(image))
    # print(numpy.max(denoised), numpy.max(image))
    denoised = denoised.clip(0, 1)
    ssim_value = ssim(denoised, image)
    psnr_value = psnr(denoised, image)

    print("denoised", psnr_value, ssim_value)

    assert ssim_value > min_ssim



# flake8: noqa
import numpy
from numpy.linalg import norm
from scipy.ndimage import median_filter
from skimage.util import random_noise

from aydin.io.datasets import camera, normalise, newyork
from aydin.it.transforms.salt_pepper import SaltPepperTransform


def test_suppress_fixed_background_real():
    image = normalise(newyork())
    noisy = random_noise(image, mode="s&p", amount=0.03, seed=0, clip=False)

    bpc = SaltPepperTransform(threshold=0.15)

    corrected = bpc.preprocess(noisy)

    median = median_filter(image, size=3)

    # import napari

    # with napari.gui_qt():
    # viewer = napari.Viewer()
    # viewer.add_image(image, name='image')
    # viewer.add_image(median, name='median')
    # viewer.add_image(corrected, name='corrected')

    error0 = numpy.abs(median - image).mean()
    error = numpy.abs(corrected - image).mean()

    print(f"Error noisy = {error0}")
    print(f"Error = {error}")
    assert error < 0.03
    assert error0 > error



import numpy

from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import lsection, lprint


class RFRegressor(LGBMRegressor):
    """Random Forrest Regressor (uses the LGBM library)."""

    def __init__(
        self,
        num_leaves: int = 1024,
        max_num_estimators: int = 2048,
        max_bin: int = 512,
        learning_rate: float = 0.0001,
        loss: str = 'l1',
        patience: int = 32,
        verbosity: int = 100,
    ):
        """Constructs a Random Forest regressor

        Parameters
        ----------
        num_leaves
            Number of leaves
        max_num_estimators
            Maximum number of estimators
        max_bin
            Maximum number of allowed bins
        learning_rate
            Learning rate for the LightGBM model
        loss
            Type of loss to be used
        patience
            Number of rounds required for early stopping
        verbosity

        """
        super().__init__(
            num_leaves,
            max_num_estimators,
            max_bin,
            learning_rate,
            loss,
            patience,
            verbosity,
        )

        with lsection("Random Forest Regressor"):
            lprint("with no arguments")  # TODO: fix these logs

    def _get_params(self, num_samples, dtype=numpy.float32):
        params = super()._get_params(num_samples, dtype)
        params["boosting_type"] = "rf"
        params["feature_fraction"] = 0.8
        return params



# flake8: noqa
from scipy.ndimage import median_filter

from aydin.io.datasets import (
    normalise,
    newyork,
    characters,
    pollen,
    lizard,
    dots,
    camera,
    add_noise,
)
from aydin.it.transforms.salt_pepper import SaltPepperTransform
from aydin.util.log.log import Log


def demo_salt_and_pepper(image=newyork()):
    Log.override_test_exclusion = True
    Log.enable_output = True

    image = normalise(image)
    noisy = add_noise(image, intensity=64, variance=0.001, sap=0.1)
    noisy = add_noise(noisy, intensity=128, variance=0.01, sap=0)
    noisy = add_noise(noisy, intensity=256, variance=0.001, sap=0.1)

    bpc = SaltPepperTransform()

    corrected = bpc.preprocess(noisy)

    median = median_filter(noisy, size=3)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(median, name='median_filtered')
        viewer.add_image(corrected, name='corrected')


demo_salt_and_pepper(newyork())
demo_salt_and_pepper(characters())
demo_salt_and_pepper(pollen())
demo_salt_and_pepper(lizard())
demo_salt_and_pepper(dots())
demo_salt_and_pepper(camera())



# flake8: noqa
import napari
import numpy as np
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import add_noise, newyork, normalise
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor


def demo(image, max_level=3):
    image = normalise(image.astype(np.float32))
    noisy = add_noise(image)

    # run wsf:
    generator = StandardFeatureGenerator(
        include_spatial_features=True, max_level=max_level
    )
    regressor = CBRegressor(patience=30)
    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)
    it.train(noisy, noisy)
    denoised_wsf = it.translate(noisy)

    # run wsf on flipped image
    flipped_img = np.flipud(noisy)
    denoised_wsf_flipped = np.flipud(it.translate(flipped_img))

    # run:
    generator = StandardFeatureGenerator(max_level=max_level)
    regressor = CBRegressor(patience=30)
    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)
    it.train(noisy, noisy)
    denoised = it.translate(noisy)

    # stats
    image_clipped = np.clip(image, 0, 1)
    noisy_clipped = np.clip(noisy, 0, 1)
    psnr_noisy = psnr(image_clipped, noisy_clipped)
    ssim_noisy = ssim(image_clipped, noisy_clipped)

    denoised_wsf_clipped = np.clip(denoised_wsf, 0, 1)
    psnr_denoised_wsf_clipped = psnr(image_clipped, denoised_wsf_clipped)
    ssim_denoised_wsf_clipped = ssim(image_clipped, denoised_wsf_clipped)

    denoised_clipped = np.clip(denoised, 0, 1)
    psnr_denoised_clipped = psnr(image_clipped, denoised_clipped)
    ssim_denoised_clipped = ssim(image_clipped, denoised_clipped)

    denoised_wsf_flipped_clipped = np.clip(denoised_wsf_flipped, 0, 1)
    psnr_denoised_wsf_flipped_clipped = psnr(
        image_clipped, denoised_wsf_flipped_clipped
    )
    ssim_denoised_wsf_flipped_clipped = ssim(
        image_clipped, denoised_wsf_flipped_clipped
    )

    print("noisy   :", psnr_noisy, ssim_noisy)
    print("denoised_wsf:", psnr_denoised_wsf_clipped, ssim_denoised_wsf_clipped)
    print("denoised:", psnr_denoised_clipped, ssim_denoised_clipped)
    print(
        "denoised_wsf_flipped:",
        psnr_denoised_wsf_flipped_clipped,
        ssim_denoised_wsf_flipped_clipped,
    )

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(denoised_wsf, name='denoised_with_sf')
        viewer.add_image(denoised, name='denoised_without_sf')
        viewer.add_image(denoised_wsf_flipped, name='denoised_with_sf_flipped')

        # imsave("ss5_wsf.tif", denoised_wsf)
        # imsave("ss5.tif", denoised)


if __name__ == '__main__':
    np.random.seed(0)
    # image = camera()
    # demo(image)
    image = newyork()
    demo(image)



import numpy
import pytest

from aydin.analysis.image_metrics import (
    mutual_information,
    joint_information,
    spectral_mutual_information,
    spectral_psnr,
)
from aydin.io.datasets import camera, add_noise, normalise


def test_spectral_psnr():
    camera_image = normalise(camera()).astype(numpy.float)
    camera_image_with_noise_high = add_noise(camera_image)
    camera_image_with_noise_low = add_noise(
        camera_image, intensity=10, variance=0.1, sap=0.000001
    )

    ji_high = spectral_psnr(camera_image, camera_image_with_noise_high)
    ji_low = spectral_psnr(camera_image, camera_image_with_noise_low)

    assert ji_high > ji_low


def test_mutual_information():
    camera_image = camera()
    camera_image_with_noise = add_noise(camera())

    mi = mutual_information(camera_image, camera_image, normalised=False)
    mi_n = mutual_information(camera_image, camera_image_with_noise, normalised=False)

    assert mi > mi_n


def test_normalised_mutual_information():
    camera_image = camera()
    camera_image_with_noise = add_noise(camera())

    assert pytest.approx(
        mutual_information(camera_image, camera_image, normalised=True), 1
    )
    assert pytest.approx(
        mutual_information(
            camera_image_with_noise, camera_image_with_noise, normalised=True
        ),
        1,
    )

    assert (
        mutual_information(camera_image, camera_image_with_noise, normalised=True) < 1
    )


def test_spectral_mutual_information():
    camera_image = camera()
    camera_image_with_noise = add_noise(camera())

    smi = spectral_mutual_information(camera_image, camera_image)
    smi_n = spectral_mutual_information(camera_image, camera_image_with_noise)

    assert smi_n < smi


def test_joint_information():
    camera_image = camera()
    camera_image_with_noise = add_noise(camera(), intensity=5, variance=3)

    ji = joint_information(camera_image, camera_image)
    ji_n = joint_information(camera_image, camera_image_with_noise)

    assert ji < ji_n



import os
import zipfile
from enum import Enum
from os.path import join, exists
from typing import Optional

import gdown
import numpy
import skimage
from scipy.ndimage import binary_dilation, zoom
from scipy.signal import convolve
from scipy.signal import convolve2d
from skimage.exposure import rescale_intensity
from skimage.util import random_noise

from aydin.io import io
from aydin.io.folders import get_cache_folder
from aydin.util.log.log import lprint
from aydin.util.psf.simple_microscope_psf import SimpleMicroscopePSF

datasets_folder = join(get_cache_folder(), 'data')

try:
    os.makedirs(datasets_folder)
except Exception:
    pass


def normalise(image):
    return rescale_intensity(
        image.astype(numpy.float32, copy=False), in_range='image', out_range=(0, 1)
    )


def add_noise(
    image,
    intensity=5,
    variance=0.01,
    sap=0.0,
    dtype=numpy.float32,
    clip=True,
    seed: Optional[int] = None,
):
    if seed is not None:
        seed = abs(seed)
        numpy.random.seed(seed)
    noisy = image
    if intensity is not None:
        noisy = numpy.random.poisson(image * intensity) / intensity
    noisy = random_noise(noisy, mode="gaussian", var=variance, seed=seed, clip=clip)
    noisy = random_noise(noisy, mode="s&p", amount=sap, seed=seed, clip=clip)
    noisy = noisy.astype(dtype, copy=False)
    return noisy


def add_blur_2d(image, dz=0):
    psf = SimpleMicroscopePSF()
    psf_xyz_array = psf.generate_xyz_psf(dxy=0.406, dz=0.406, xy_size=17, z_size=17)
    psf_kernel = psf_xyz_array[dz]
    psf_kernel /= psf_kernel.sum()
    return convolve2d(image, psf_kernel, mode='same', boundary="symm"), psf_kernel


def add_blur_3d(image, xy_size=17, z_size=17):
    psf = SimpleMicroscopePSF()
    psf_xyz_array = psf.generate_xyz_psf(
        dxy=0.406, dz=0.406, xy_size=xy_size, z_size=z_size
    )
    psf_kernel = psf_xyz_array
    psf_kernel /= psf_kernel.sum()
    return convolve(image, psf_kernel, mode='same'), psf_kernel


# Convenience shortcuts:


def lizard():
    return examples_single.generic_lizard.get_array()


def camera():
    return skimage.data.camera().astype(numpy.float32, copy=False)


def newyork():
    return examples_single.generic_newyork.get_array()


def small_newyork():
    return zoom(newyork(), zoom=0.5)


def cropped_newyork(crop_amount=256):
    crop_amount = min(crop_amount, 500)
    return newyork()[crop_amount:-crop_amount, crop_amount:-crop_amount]


def pollen():
    return examples_single.generic_pollen.get_array()


def scafoldings():
    return examples_single.generic_scafoldings.get_array()


def characters():
    return 1 - examples_single.generic_characters.get_array()


def andromeda():
    return examples_single.generic_andromeda.get_array()


def dots():
    image = numpy.random.rand(512, 512) < 0.005  # andromeda()#[256:-256, 256:-256]
    image = 0.8 * binary_dilation(image).astype(numpy.float32, copy=False)
    image[0:256, 0:256] += 0.1
    image.clip(0, 1)
    return image


def rgbtest():
    return examples_single.rgbtest.get_array()


def dmel():
    return examples_single.keller_dmel.get_array()[23]


class examples_single(Enum):
    def get_path(self):
        download_from_gdrive(*self.value, datasets_folder)
        return join(datasets_folder, self.value[1])

    def get_array(self):
        array, _ = io.imread(self.get_path())
        return array

    fountain = ('1JP-_j-6U7J1gNc9IZCZ_GsgXTcybmZgS', 'fountain.png')
    monalisa = ('15T3oTCyz7ugnPLTsKc0a9NT17g9GJsO_', 'monalisa.png')
    gauss_noisy = ('17e_ECJA7DUQGu9JELbTkAKbOVVE9olHN', 'Gauss_noisy.png')
    periodic_noise = ('1HfwF6gnzHFFdJ-tozllU_h14vNk9GZOG', 'periodic_noise.png')
    brown_chessboard_gray = (
        '1gnqwhZ7HrRaScj6QF_P2Pl_6WAcLzCgR',
        'Brown_SIDD_chessboard_gray.png',
    )

    # XY natural images (2D monochrome):
    generic_crowd = ('13UHK8MjhBviv31mAW2isdG4G-aGaNJIj', 'crowd.tif')
    generic_mandrill = ('1B33ELiFuCV0OJ6IHh7Ix9lvImwI_QkR-', 'mandrill.tif')
    generic_newyork = ('15Nuu_NU3iNuoPRmpFbrGIY0VT0iCmuKu', 'newyork.png')
    generic_lizard = ('1GUc6jy5QH5DaiUskCrPrf64YBOLzT6j1', 'lizard.png')
    generic_pollen = ('1S0o2NWtD1shB5DfGRIqOFxTLOi8cHQD-', 'pollen.png')
    generic_scafoldings = ('1ZiWhHnkuaQH-BS8B71y00wkN1Ylo38nY', 'scafoldings.png')
    generic_andromeda = ('1Zl3DtkwUlZSbvpxGILexiIoLW1JOdJh8', 'andromeda.png')

    # Characters (2D monochrome, inverted):
    generic_characters = ('1ZWkHFI2iddKa9qv6tft4QZlCoDS5fLMK', 'characters.jpg')

    # XYC (RGB)
    rgbtest = ('1KvhcGBqEQ5N9mwxHwy14NVp8OJ-9GCsH', 'rgbtest.png')

    leonetti_tm7sf2 = (
        '1HHsbZ6jyuJkIj6c7kGtsPKOgpUxo0ihw',
        'Leonetti_p4B3_1_TM7SF2_PyProcessed_IJClean.tif',
    )
    leonetti_sptssa = (
        '10kR7FSIyi7417XYTLrMJaGe3MfvMmSYA',
        'Leonetti_p1H10_2_SPTSSA_PyProcessed_IJClean.tif',
    )
    leonetti_snca = (
        '1UyF5HkZLwTaoiBf1sLHkTdw09yyCJyKO',
        'Leonetti_p1H8_2_SNCA_PyProcessed_IJClean.tif',
    )

    # XYZ
    huang_fixed_pattern_noise = (
        '1fiVopWwUSZJhsWUfoo7YaU3qQ-j4g5hB',
        'fixed_pattern_noise.tif',
    )

    keller_dmel = (
        '12DCAlDRSiTyGDSD7p06nk17GO3ztHg-Q',
        'SPC0_TM0132_CM0_CM1_CHN00_CHN01.fusedStack.tif',
    )

    janelia_flybrain = (
        '12Z6W_f3TqCsl_okKmaLcVUBgS6xEvdjj',
        'Flybrain_3ch_mediumSize.tif',
    )

    myers_tribolium = (
        '1GGULAF8IoPi4P614Wi-njmeQxND4Xttj',
        'Myers_Tribolium_nGFP_0.1_0.2_0.5_20_13_late.tif',
    )

    royerlab_hcr = (
        '1nkgqs8VkmPKBtBHqUXKrVgIvOqD8B84h',
        'Royer_confocal_dragonfly_hcr_drerio_30somite_crop.tif',
    )

    # 2D+t

    cognet_nanotube1 = (
        '1SmrBheUc6p5qTgtIEzedCwbN87HOW_O_',
        'Cognet_r03-s01-100mW-20ms-175 50xplpeg-173.tif',
    )
    cognet_nanotube_400fps = (
        '1ap4YNxa0RA6MBKiXZ2ZRL1USZtzPqFs3',
        'Cognet_1-400fps.tif',
    )
    cognet_nanotube_200fps = (
        '1Z501FlQOBQmPaeBMCOGy6chBDh1bDjEf',
        'Cognet_1-200fps.tif',
    )
    cognet_nanotube_100fps = (
        '1T4UvbF3MRgT4jO4ExIHprvTqUXLiMjyA',
        'Cognet_1-100fps.tif',
    )

    # 3D+t
    maitre_mouse = (
        '13b0-6PUo2YEWG8Z3M1pVfBQBWE0PtILK',
        'Maitre_mouse blastocyst_fracking_180124_e3_crop.tif',
    )

    # XYZT
    hyman_hela = ('12qOGxfBrnzrufgbizyTkhHipgRwjSIz-', 'Hyman_HeLa.tif')

    # XYZCT 1344 × 1024 × 1 × 1 × 93
    ome_mitocheck = ('1B9d8Yw_lidZg43U3VZAoalVHf9eHbCS7', '00001_01.ome.tiff')

    # XYZCT 160 × 220 × 8 × 2 × 12
    ome_spim = ('1BG6jCZGLEs1LDxKXjMqF0aV-iiqlushk', 'SPIM-ModuloAlongZ.ome.tiff')


def download_from_gdrive(
    id, name, dest_folder=datasets_folder, overwrite=False, unzip=False
):

    try:
        os.makedirs(dest_folder)
    except Exception:
        pass

    url = f'https://drive.google.com/uc?id={id}'
    output_path = join(dest_folder, name)
    if overwrite or not exists(output_path):
        lprint(f"Downloading file {output_path} as it does not exist yet.")
        gdown.download(url, output_path, quiet=False)

        if unzip:
            lprint(f"Unzipping file {output_path}...")
            zip_ref = zipfile.ZipFile(output_path, 'r')
            zip_ref.extractall(dest_folder)
            zip_ref.close()
            # os.remove(output_path)

        return output_path
    else:
        lprint(f"Not downloading file {output_path} as it already exists.")
        return None


def download_all_examples():
    for example in examples_single:
        print(download_from_gdrive(*example.value))


def downloaded_example(substring):
    for example in examples_single.get_list():
        if substring in example.value[1]:
            print(download_from_gdrive(*example.value))



import re

import numpy
from m2cgen import interpreters
from pyopencl.array import Array

from aydin.providers.opencl.opencl_provider import OpenCLProvider
from aydin.regression.gbm_utils.light_gbm_assembler import LightGBMModelAssembler
from aydin.util.array.nd import nd_split_slices
from aydin.util.log.log import lsection, lprint
from aydin.util.misc.recursion import RecursionLimit


class GBMOpenCLPrediction:
    """
    Class that handles GBM OpenCL prediction.

    """

    def __init__(self, num_trees_per_kernel_call=100, max_gpu_length_array=6 * 1e6):

        """
        Constructs a GBM OpenCL inference instance
        """

        super().__init__()

        self.num_trees_per_kernel_call = num_trees_per_kernel_call
        self.max_gpu_length_array = max_gpu_length_array

        self.opencl_provider: OpenCLProvider = None
        self.x_gpu = None
        self.y_gpu = None

    def _ensure_opencl_prodider_initialised(self):
        if not hasattr(self, 'opencl_provider') or self.opencl_provider is None:
            self.opencl_provider = OpenCLProvider()

    def export_to_opencl_c(
        self, model, function_name='score', dtype='float', indent=4, tree_slice=None
    ):

        code = self.export_to_c(model, indent=indent, tree_slice=tree_slice)

        code = code.replace('double score(', 'inline double score(__global ')

        code = code.replace('score(', function_name + '(')

        code = code.replace('double', dtype)

        if dtype == 'float':
            code = re.sub(r'([\d]+\.[\d]+[-+e\d]*)', r'\1f', code)

        return code

    def export_to_c(self, model, indent=4, tree_slice=None):

        with RecursionLimit(3000):
            interpreter = interpreters.CInterpreter(indent=indent)
            model_ast = LightGBMModelAssembler(model, tree_slice=tree_slice).assemble()
            code = interpreter.interpret(model_ast)
            return code

    def predict(self, model, x, num_iteration=None, dtype='float', indent=4):

        with lsection("GBM OpenCL prediction:"):
            self._ensure_opencl_prodider_initialised()

            num_features = x.shape[-1]
            num_datapoints = x.shape[0]

            def get_program(tree_slice=slice(None)):
                forrest_code = self.export_to_opencl_c(
                    model, dtype=dtype, indent=indent, tree_slice=tree_slice
                )

                return f"""
                 {forrest_code}
                 __kernel void gbm_kernel(__global float *x, __global float *y)
                 {{
                   const int i = get_global_id(0);
                   const int x_offset = {num_features}*i ;
                   const int y_offset = i ;
                   y[y_offset] = score(x+x_offset);
                 }}
                 """

            x = numpy.ascontiguousarray(x, dtype=numpy.float32)

            gpu_array_length = min(int(self.max_gpu_length_array), num_datapoints)
            lprint(f"Proposed GPU array length: {gpu_array_length}")

            nb_array_slices = max(1, num_datapoints // gpu_array_length)
            array_slices = list(
                s[0] for s in nd_split_slices((num_datapoints,), (nb_array_slices,))
            )
            gpu_array_length = array_slices[0].stop - array_slices[0].start
            lprint(f"Effective GPU array length: {gpu_array_length}")

            if self.x_gpu is None or (
                self.x_gpu.shape != (gpu_array_length, num_features)
                and self.x_gpu.dtype != numpy.float32
            ):
                self.x_gpu = Array(
                    self.opencl_provider.queue,
                    (gpu_array_length, num_features),
                    numpy.float32,
                )

            if self.y_gpu is None or (
                self.y_gpu.shape != (gpu_array_length,)
                and self.y_gpu.dtype != numpy.float32
            ):
                self.y_gpu = Array(
                    self.opencl_provider.queue, (gpu_array_length,), numpy.float32
                )
                self.y_gpu.fill(0)

            num_trees = num_iteration
            nb_kernel_slices = max(1, num_trees // self.num_trees_per_kernel_call)
            kernel_slices = list(
                s[0] for s in nd_split_slices((num_iteration,), (nb_kernel_slices,))
            )

            lprint(f"Number of trees                : {num_trees}")
            lprint(f"Number of trees per kernel call: {self.num_trees_per_kernel_call}")
            lprint(f"Number of kernel slices        : {nb_kernel_slices}")

            y = numpy.zeros(num_datapoints)

            with lsection("Running kernels:"):
                for index, one_slice in enumerate(kernel_slices):
                    with lsection(
                        f"Kernel slice {index+1}/{len(kernel_slices)}: {one_slice}"
                    ):
                        with lsection("Generating kernel program code:"):
                            program_code = get_program(tree_slice=one_slice)
                            lprint(f"Kernel program length: {len(program_code)}")

                        program = self.opencl_provider.build(program_code)
                        gbm_kernel = program.gbm_kernel

                        with lsection(
                            f"Running kernel for {len(array_slices)} array slices"
                        ):
                            for one_array_slice in array_slices:
                                x_slice = x[one_array_slice]
                                x_slice = numpy.pad(
                                    x_slice,
                                    pad_width=(
                                        (0, self.x_gpu.shape[0] - x_slice.shape[0]),
                                        (0, 0),
                                    ),
                                    mode='constant',
                                )
                                self.x_gpu.set(x_slice, self.opencl_provider.queue)
                                # self.y_gpu.fill(0)

                                with lsection(
                                    f"Running kernel for array slice : {one_array_slice}"
                                ):
                                    gbm_kernel(
                                        self.opencl_provider.queue,
                                        (gpu_array_length,),
                                        None,
                                        self.x_gpu.data,
                                        self.y_gpu.data,
                                    )

                                    subarray_length = (
                                        one_array_slice.stop - one_array_slice.start
                                    )
                                    y[one_array_slice] += self.y_gpu.get(
                                        self.opencl_provider.queue
                                    )[0:subarray_length]

            del self.x_gpu
            del self.y_gpu

            return y
